# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

from functools import singledispatch

import numpy as np
import pandas as pd
import torch
from pyro.ops.stats import pi, resample
from pyroapi import distributions as dist


@singledispatch
def count_intervals(labels):
    r"""
    Count binding interval data.

    Co-localization and absent intervals are coded as -3
    and -2 respectively when they are the first (or only) interval in a record, 3 and 2
    when they are the last interval in a record and 1 and 0 elsewhere.

    Reference::

      @article{friedman2015multi,
        title={Multi-wavelength single-molecule fluorescence analysis of transcription mechanisms},
        author={Friedman, Larry J and Gelles, Jeff},
        journal={Methods},
        volume={86},
        pages={27--36},
        year={2015},
        publisher={Elsevier}
      }
    """
    raise NotImplementedError


@count_intervals.register(np.ndarray)
def _(labels):
    labels = labels.astype("bool")
    start_condition = (
        np.concatenate((~labels[:, 0:1], labels[:, :-1]), axis=1) != labels
    )
    start_aoi, start_frame = np.nonzero(start_condition)
    start_type = labels.astype("long")
    start_type[:, 0] = -start_type[:, 0] - 2
    start_type = start_type[start_aoi, start_frame]

    stop_condition = np.concatenate(
        (labels[:, :-1] != labels[:, 1:], np.ones_like(labels[:, 0:1])), axis=1
    )
    stop_aoi, stop_frame = np.nonzero(stop_condition)
    stop_type = labels.astype("long")
    stop_type[:, -1] += 2
    stop_type = stop_type[stop_aoi, stop_frame]

    assert all(start_aoi == stop_aoi)

    low_or_high = np.where(abs(start_type) > abs(stop_type), start_type, stop_type)

    result = pd.DataFrame(
        data={
            "aoi": start_aoi,
            "start_frame": start_frame,
            "stop_frame": stop_frame,
            "dwell_time": stop_frame + 1 - start_frame,
            "low_or_high": low_or_high,
        }
    )
    return result


@count_intervals.register(torch.Tensor)
def _(labels):
    labels = labels.bool()
    start_condition = torch.cat((~labels[:, 0:1], labels[:, :-1]), dim=1) != labels
    start_aoi, start_frame = torch.nonzero(start_condition, as_tuple=True)
    start_type = labels.long()
    start_type[:, 0] = -start_type[:, 0] - 2
    start_type = start_type[start_aoi, start_frame]

    stop_condition = torch.cat(
        (labels[:, :-1] != labels[:, 1:], torch.ones_like(labels[:, 0:1])), dim=1
    )
    stop_aoi, stop_frame = torch.nonzero(stop_condition, as_tuple=True)
    stop_type = labels.long()
    stop_type[:, -1] += 2
    stop_type = stop_type[stop_aoi, stop_frame]

    assert all(start_aoi == stop_aoi)

    low_or_high = torch.where(abs(start_type) > abs(stop_type), start_type, stop_type)

    result = pd.DataFrame(
        data={
            "aoi": start_aoi,
            "start_frame": start_frame,
            "stop_frame": stop_frame,
            "dwell_time": stop_frame + 1 - start_frame,
            "low_or_high": low_or_high,
        }
    )
    return result


def bound_dwell_times(intervals):
    assert isinstance(intervals, pd.DataFrame)
    mask = intervals["low_or_high"] == 1
    return intervals.loc[mask, "dwell_time"].values


def unbound_dwell_times(intervals):
    assert isinstance(intervals, pd.DataFrame)
    mask = intervals["low_or_high"] == 0
    return intervals.loc[mask, "dwell_time"].values


@singledispatch
def time_to_first_binding(labels):
    r"""
    Measure the time elapsed prior to the first binding.

    Time-to-first binding for a binary data:

        :math:`\mathrm{ttfb} =
        \sum_{f=1}^{F-1} f z_{n,f} \prod_{f^\prime=0}^{f-1} (1 - z_{n,f^\prime})
        + F \prod_{f^\prime=0}^{F-1} (1 - z_{n,f^\prime})`

    Expected value of the time-to-first binding:

        :math:`\mathbb{E}[\mathrm{ttfb}] =
        \sum_{f=1}^{F-1} f q(z_{n,f}=1) \prod_{f^\prime=f-1}^{f-1} q(z_{n,f^\prime}=0)
        + F \prod_{f^\prime=0}^{F-1} q(z_{n,f^\prime}=0)`

    Reference::

      @article{friedman2015multi,
        title={Multi-wavelength single-molecule fluorescence analysis of transcription mechanisms},
        author={Friedman, Larry J and Gelles, Jeff},
        journal={Methods},
        volume={86},
        pages={27--36},
        year={2015},
        publisher={Elsevier}
      }
    """
    raise NotImplementedError


@time_to_first_binding.register(np.ndarray)
def _(labels):
    labels = labels.astype("float")
    F = labels.shape[-1]
    frames = np.arange(1, F + 1)
    q1 = np.ones_like(labels)
    q1[..., :-1] = labels[..., 1:]
    cumq0 = np.cumprod(1 - labels, axis=-1)
    ttfb = (frames * q1 * cumq0).sum(-1)
    return ttfb


@time_to_first_binding.register(torch.Tensor)
def _(labels):
    labels = labels.float()
    F = labels.shape[-1]
    frames = torch.arange(1, F + 1)
    q1 = torch.ones_like(labels)
    q1[..., :-1] = labels[..., 1:]
    cumq0 = torch.cumprod(1 - labels, dim=-1)
    ttfb = (frames * q1 * cumq0).sum(-1)
    return ttfb


@singledispatch
def association_rate(labels):
    r"""
    Compute the on-rate from the binary data assuming a two-state HMM model.
    """
    raise NotImplementedError


@association_rate.register(np.ndarray)
def _(labels):
    binding_events = ((1 - labels[..., :-1]) * labels[..., 1:]).sum((-2, -1))
    off_states = (1 - labels[..., :-1]).sum((-2, -1))
    kon = binding_events / off_states
    return kon


@association_rate.register(torch.Tensor)
def _(labels):
    labels = labels.float()
    binding_events = ((1 - labels[..., :-1]) * labels[..., 1:]).sum((-2, -1))
    off_states = (1 - labels[..., :-1]).sum((-2, -1))
    kon = binding_events / off_states
    return kon


@singledispatch
def dissociation_rate(labels):
    r"""
    Compute the off-rate from the binary data assuming a two-state HMM model.
    """
    raise NotImplementedError


@dissociation_rate.register(np.ndarray)
def _(labels):
    dissociation_events = (labels[..., :-1] * (1 - labels[..., 1:])).sum((-2, -1))
    on_states = labels[..., :-1].sum((-2, -1))
    koff = dissociation_events / on_states
    return koff


@dissociation_rate.register(torch.Tensor)
def _(labels):
    labels = labels.float()
    dissociation_events = (labels[..., :-1] * (1 - labels[..., 1:])).sum((-2, -1))
    on_states = labels[..., :-1].sum((-2, -1))
    koff = dissociation_events / on_states
    return koff


@singledispatch
def bootstrap(samples, estimator, repetitions=1000, probs=0.68):
    r"""
    Estimate the confidence interval of an estimator by constructing approximating
    distributions using the bootstrap method (resampling with replacement).
    """
    raise NotImplementedError


@bootstrap.register(np.ndarray)
def _(samples, estimator, repetitions=1000, probs=0.68):
    estimand = np.zeros((repetitions,))
    for i in range(repetitions):
        bootstrap_values = np.random.choice(samples, size=len(samples), replace=True)
        estimand[i] = estimator(bootstrap_values)
    return np.quantile(estimand, (1 - probs) / 2), np.quantile(
        estimand, (1 + probs) / 2
    )


@bootstrap.register(torch.Tensor)
def _(samples, estimator, repetitions=1000, probs=0.68):
    estimand = torch.zeros(repetitions)
    for i in range(repetitions):
        bootstrap_values = resample(samples, num_samples=len(samples), replacement=True)
        estimand[i] = estimator(bootstrap_values)
    return pi(estimand, probs)


@singledispatch
def posterior_estimate(dist, estimator, repetitions=1000, probs=0.68):
    r"""
    A version of bootstrapping method where samples are first drawn from
    a distribution and then resampled with replacement.
    """
    raise NotImplementedError


@posterior_estimate.register(dist.Distribution)
def _(dist, estimator, repetitions=1000, probs=0.68):
    samples = dist.sample((repetitions,))
    estimand = torch.zeros(repetitions)
    for i in range(repetitions):
        estimand[i] = estimator(samples[i])
    return pi(estimand, probs)


@singledispatch
def sample_and_bootstrap(
    dist, estimator, preprocess=None, repetitions=1000, probs=0.68
):
    r"""
    A version of bootstrapping method where samples are first drawn from
    a distribution and then resampled with replacement.
    """
    raise NotImplementedError


@sample_and_bootstrap.register(dist.Distribution)
def _(dist, estimator, preprocess=None, repetitions=1000, probs=0.68):
    estimand = torch.zeros(repetitions)
    for i in range(repetitions):
        samples = dist.sample()
        if preprocess is not None:
            samples = preprocess(samples)
        # bootstrap_values = resample(samples, num_samples=len(samples), replacement=True)
        bootstrap_values = np.random.choice(samples, size=len(samples), replace=True)
        estimand[i] = estimator(bootstrap_values)
    return pi(estimand, probs)



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import pytest
import torch
from typer.testing import CliRunner

from tapqir.main import app
from tapqir.models import HMM, Cosmos
from tapqir.utils.dataset import save
from tapqir.utils.simulate import simulate

requires_cuda = pytest.mark.skipif(
    not torch.cuda.is_available(), reason="cuda is not available"
)

runner = CliRunner()


@pytest.fixture(params=[Cosmos, HMM])
def dataset_path(request, tmp_path):
    params = {}
    if request.param == Cosmos:
        model = request.param()
        params["pi"] = 0.15
    elif request.param == HMM:
        model = request.param(vectorized=False)
        params["kon"] = 0.2
        params["koff"] = 0.2
    params["width"] = 1.4
    params["gain"] = 7.0
    params["lamda"] = 0.15
    params["proximity"] = 0.2
    params["offset"] = 90.0
    params["height"] = 3000
    params["background"] = 150
    N = 2
    F = 5
    C = 1
    P = 14

    data = simulate(model, N, F, C, P, params=params)

    # save data
    save(data, tmp_path)
    return tmp_path


def test_commands_cpu(dataset_path):

    commands = [
        [
            "--cd",
            dataset_path,
            "fit",
            "--model",
            "cosmos",
            "--learning-rate",
            "0.005",
            "--nbatch-size",
            "2",
            "--fbatch-size",
            "5",
            "--num-iter",
            "1",
            "--cpu",
            "--no-input",
        ],
        [
            "--cd",
            dataset_path,
            "stats",
            "--model",
            "cosmos",
            "--nbatch-size",
            "2",
            "--fbatch-size",
            "5",
            "--cpu",
            "--no-input",
        ],
    ]

    for command in commands:
        result = runner.invoke(app, command)
        assert result.exit_code == 0


@requires_cuda
def test_commands_cuda(dataset_path):
    commands = [
        [
            "--cd",
            dataset_path,
            "fit",
            "--model",
            "cosmos",
            "--learning-rate",
            "0.005",
            "--cuda",
            "--nbatch-size",
            "2",
            "--fbatch-size",
            "5",
            "--num-iter",
            "1",
            "--no-input",
        ],
    ]

    for command in commands:
        result = runner.invoke(app, command)
        assert result.exit_code == 0



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import torch
from matplotlib.patches import Rectangle
from scipy.io import loadmat

from tapqir.utils.dataset import CosmosDataset, save

# logger
logger = logging.getLogger(__name__)


def bin_hist(
    samples: torch.Tensor, weights: torch.Tensor, s: int
) -> Tuple[torch.Tensor, torch.Tensor]:
    q, r = divmod(len(samples) - 1, s)
    n = 1 + q + int(bool(r))
    new_samples = torch.zeros(n, dtype=torch.int)
    new_weights = torch.zeros(n)
    new_samples[0] = samples[0]
    new_weights[0] = weights[0]
    new_samples[1 : 1 + q] = samples[1 + s // 2 : 1 + q * s : s]
    for i in range(s):
        new_weights[1 : 1 + q] += weights[1 + i : 1 + q * s : s]
    if r:
        new_samples[-1] = samples[1 + q * s + r // 2]
        new_weights[-1] = weights[1 + q * s :].sum()
    return new_samples, new_weights


class GlimpseDataset:
    """
    GlimpseDataset parses header, aoiinfo, driftlist, and intervals (optional)
    files.

    :param name: Channel name.
    :param glimpse-folder: Path to the header/glimpse folder.
    :param ontarget-aoiinfo: Path to the on-target AOI locations file.
    :param offtarget-aoiinfo: Path to the off-target control AOI locations file (optional).
    :param driftlist: Path to the driftlist file.
    :param frame-start: First frame to include in the analysis (optional).
    :param frame-end: Last frame to include in the analysis (optional).
    :param ontarget-labels: Path to the on-target label intervals file.
    """

    def __init__(self, c=0, **kwargs):
        dtypes = ["ontarget"]
        if kwargs["use-offtarget"]:
            dtypes.append("offtarget")

        # convert header into dict format
        mat_header = loadmat(Path(kwargs["glimpse-folder"]) / "header.mat")
        header = dict()
        for i, dt in enumerate(mat_header["vid"].dtype.names):
            header[dt] = np.squeeze(mat_header["vid"][0, 0][i])

        # load driftlist mat file
        drift_mat = loadmat(kwargs["driftlist"])
        # convert driftlist into DataFrame
        drift_df = pd.DataFrame(
            drift_mat["driftlist"][:, :3], columns=["frame", "dy", "dx"]
        )
        drift_df = drift_df.astype({"frame": int}).set_index("frame")
        drift_df["ttb"] = header["ttb"]

        # load aoiinfo mat file
        aoi_mat = {}
        aoi_df = {}
        for dtype in dtypes:
            try:
                aoi_mat[dtype] = loadmat(kwargs[f"{dtype}-aoiinfo"])
            except ValueError:
                aoi_mat[dtype] = np.loadtxt(kwargs[f"{dtype}-aoiinfo"])
            try:
                aoi_df[dtype] = pd.DataFrame(
                    aoi_mat[dtype]["aoiinfo2"],
                    columns=["frame", "ave", "y", "x", "pixnum", "aoi"],
                )
            except KeyError:
                aoi_df[dtype] = pd.DataFrame(
                    aoi_mat[dtype]["aoifits"]["aoiinfo2"][0, 0],
                    columns=["frame", "ave", "y", "x", "pixnum", "aoi"],
                )
            except IndexError:
                aoi_df[dtype] = pd.DataFrame(
                    aoi_mat[dtype], columns=["frame", "ave", "y", "x", "pixnum", "aoi"]
                )
            aoi_df[dtype] = aoi_df[dtype].astype({"aoi": int}).set_index("aoi")
            # adjust to python indexing
            aoi_df[dtype]["x"] = aoi_df[dtype]["x"] - 1
            aoi_df[dtype]["y"] = aoi_df[dtype]["y"] - 1

        # calculate the cumulative sum of dx and dy relative to the aoiinfo frame
        aoiinfo_frame = int(aoi_df["ontarget"].at[1, "frame"])
        drift_df.loc[aoiinfo_frame + 1 :] = (
            drift_df.loc[aoiinfo_frame + 1 :].cumsum(axis=0).values
        )
        drift_df.loc[aoiinfo_frame - 1 :: -1] = (
            (-drift_df.loc[aoiinfo_frame : drift_df.index[1] : -1])
            .cumsum(axis=0)
            .values
        )

        if kwargs["frame-range"]:
            f1 = int(kwargs["frame-start"])
            f2 = int(kwargs["frame-end"])
            drift_df = drift_df.loc[f1:f2]

        labels = defaultdict(lambda: None)
        for dtype in dtypes:
            if kwargs["labels"] and kwargs[f"{dtype}-labels"] is not None:
                labels_mat = loadmat(kwargs[f"{dtype}-labels"])
                labels[dtype] = np.zeros(
                    (len(aoi_df[dtype]), len(drift_df)),
                    dtype=[
                        ("aoi", int),
                        ("frame", int),
                        ("z", bool),
                        ("spotpicker", float),
                    ],
                )
                labels[dtype]["aoi"] = aoi_df[dtype].index.values.reshape(-1, 1)
                labels[dtype]["frame"] = drift_df.index.values
                spot_picker = labels_mat["Intervals"]["CumulativeIntervalArray"][0, 0]
                for sp in spot_picker:
                    aoi = int(sp[-1])
                    start = int(sp[1])
                    end = int(sp[2])
                    if sp[0] in [-2.0, 0.0, 2.0]:
                        labels[dtype]["spotpicker"][
                            (labels[dtype]["aoi"] == aoi)
                            & (labels[dtype]["frame"] >= start)
                            & (labels[dtype]["frame"] <= end)
                        ] = 0
                    elif sp[0] in [-3.0, 1.0, 3.0]:
                        labels[dtype]["spotpicker"][
                            (labels[dtype]["aoi"] == aoi)
                            & (labels[dtype]["frame"] >= start)
                            & (labels[dtype]["frame"] <= end)
                        ] = 1

                labels[dtype]["z"] = labels[dtype]["spotpicker"]

        self.height, self.width = int(header["height"]), int(header["width"])
        self.config = kwargs
        self.header = header
        self.dtypes = dtypes
        self.aoiinfo = aoi_df
        self.cumdrift = drift_df
        self.labels = labels
        self.name = kwargs["name"]
        self.c = c

    def __len__(self) -> int:
        return self.F

    def __getitem__(self, key):
        # read the entire frame image
        if isinstance(key, slice):
            imgs = []
            for frame in range(
                key.start, key.stop, 1 if key.step is None else key.step
            ):
                imgs.append(self[frame])
            return np.stack(imgs, 0)
        frame = key
        glimpse_number = self.header["filenumber"][frame - 1]
        glimpse_path = Path(self.config["glimpse-folder"]) / f"{glimpse_number}.glimpse"
        offset = self.header["offset"][frame - 1]
        with open(glimpse_path, "rb") as fid:
            fid.seek(offset)
            img = np.fromfile(fid, dtype=">i2", count=self.height * self.width).reshape(
                self.height, self.width
            )
        return img + 2**15

    @property
    def N(self) -> int:
        return len(self.aoiinfo["ontarget"])

    @property
    def Nc(self) -> int:
        if "offtarget" in self.dtypes:
            return len(self.aoiinfo["offtarget"])
        return 0

    @property
    def F(self) -> int:
        return len(self.cumdrift)

    def __repr__(self):
        return rf"{self.__class__.__name__}(N={self.N}, Nc={self.Nc}, F={self.F})"

    def __str__(self):
        return f"{self.__class__.__name__}(N={self.N}, Nc={self.Nc}, F={self.F})"

    def plot(self, dtype: str, P: int, path: Path, save: bool):
        """
        Plot AOIs in the field of view.

        :param dtype: Data type (``ontarget``, ``offtarget``, ``offset``).
        :param P: AOI size.
        :param path: Path where to save plots.
        :param save: Save plots.
        """
        colors = {}
        colors["ontarget"] = "#AA3377"
        colors["offtarget"] = "#CCBB44"
        fig = plt.figure(figsize=(10, 10 * self.height / self.width))
        ax = fig.add_subplot(1, 1, 1)
        fov = self[self.cumdrift.index[0]]
        vmin = np.percentile(fov, 1)
        vmax = np.percentile(fov, 99)
        ax.imshow(fov, vmin=vmin, vmax=vmax, cmap="gray")
        if dtype in ["ontarget", "offtarget"]:
            for aoi in self.aoiinfo[dtype].index:
                # areas of interest
                y_pos = round(self.aoiinfo[dtype].at[aoi, "y"] - 0.5 * (P - 1)) - 0.5
                x_pos = round(self.aoiinfo[dtype].at[aoi, "x"] - 0.5 * (P - 1)) - 0.5
                ax.add_patch(
                    Rectangle(
                        (x_pos, y_pos),
                        P,
                        P,
                        edgecolor=colors[dtype],
                        lw=1,
                        facecolor="none",
                    )
                )
        elif dtype == "offset":
            ax.add_patch(
                Rectangle(
                    (10, 10),
                    P,
                    P,
                    edgecolor="#CCBB44",
                    lw=1,
                    facecolor="none",
                )
            )
        ax.set_title(f"{dtype} locations for channel {self.c}", fontsize=16)
        ax.set_xlabel("x", fontsize=16)
        ax.set_ylabel("y", fontsize=16)
        if save:
            plt.savefig(path / f"{dtype}-channel{self.c}.png", dpi=300)


def read_glimpse(path, progress_bar, **kwargs):
    """
    Preprocess glimpse files.
    """
    P = kwargs.pop("P")
    C = kwargs.pop("num-channels")
    name = kwargs.pop("dataset")
    channels = kwargs.pop("channels")
    offset_x = kwargs.pop("offset_x")
    offset_y = kwargs.pop("offset_y")
    offset_P = kwargs.pop("offset_P")
    bin_size = kwargs.pop("bin_size")

    offsets = defaultdict(int)
    offset_medians = []
    # iterate over channels
    data = defaultdict(list)
    target_xy = defaultdict(list)
    labels = defaultdict(list)
    time1 = []
    ttb = []
    for c in range(C):
        glimpse = GlimpseDataset(**kwargs, **channels[c], c=c)

        raw_target_xy = {}
        colors = {}
        colors["ontarget"] = "#AA3377"
        colors["offtarget"] = "#CCBB44"
        time1.append(float(glimpse.header["time1"]))
        ttb.append(glimpse.cumdrift["ttb"].values)
        for dtype in glimpse.dtypes:
            N = len(glimpse.aoiinfo[dtype])
            F = len(glimpse.cumdrift)
            raw_target_xy[dtype] = (
                np.expand_dims(glimpse.aoiinfo[dtype][["x", "y"]].values, axis=1)
                + glimpse.cumdrift[["dx", "dy"]].values
            )
            target_xy[dtype].append(np.zeros((N, F, 2)))
            data[dtype].append(
                np.zeros(
                    (N, F, P, P),
                    dtype="int",
                )
            )
            labels[dtype].append(glimpse.labels[dtype])

            glimpse.plot(dtype, P, path=path, save=True)

        # plot offset in raw FOV images
        glimpse.plot("offset", offset_P, path=path, save=True)

        # loop through each frame
        for f, frame in enumerate(progress_bar(glimpse.cumdrift.index)):
            img = glimpse[frame]

            offset_img = img[
                offset_y : offset_y + offset_P, offset_x : offset_x + offset_P
            ]
            offset_medians.append(np.median(offset_img))
            values, counts = np.unique(offset_img, return_counts=True)
            for value, count in zip(values, counts):
                offsets[value] += count
            for dtype in glimpse.dtypes:
                # loop through each aoi
                for n, aoi in enumerate(glimpse.aoiinfo[dtype].index):
                    shiftx = round(raw_target_xy[dtype][n, f, 0] - 0.5 * (P - 1))
                    shifty = round(raw_target_xy[dtype][n, f, 1] - 0.5 * (P - 1))
                    data[dtype][c][n, f, :, :] += img[
                        shifty : shifty + P, shiftx : shiftx + P
                    ]
                    target_xy[dtype][c][n, f, 0] = (
                        raw_target_xy[dtype][n, f, 0] - shiftx
                    )
                    target_xy[dtype][c][n, f, 1] = (
                        raw_target_xy[dtype][n, f, 1] - shifty
                    )

        # assert that target positions are within central pixel
        for dtype in glimpse.dtypes:
            assert (target_xy[dtype][c] > 0.5 * P - 1).all()
            assert (target_xy[dtype][c] < 0.5 * P).all()

    min_data = np.inf
    for dtype in data.keys():
        # concatenate color channels
        data[dtype] = np.stack(data[dtype], -3)
        target_xy[dtype] = np.stack(target_xy[dtype], -2)
        min_data = min(min_data, data[dtype].min())
        if any(label is None for label in labels[dtype]):
            labels[dtype] = None
        else:
            labels[dtype] = np.stack(labels[dtype], -1)
        # convert data to torch tensor
        data[dtype] = torch.tensor(data[dtype])
        target_xy[dtype] = torch.tensor(target_xy[dtype])

    # process offset data
    offsets = OrderedDict(sorted(offsets.items()))
    offset_samples = np.array(list(offsets.keys()))
    offset_weights = np.array(list(offsets.values()))
    # if data.min() is smaller than the smallest offset then
    # add a single point to offset samples with that value - 1
    if min_data <= offset_samples[0]:
        offset_samples = np.insert(offset_samples, 0, min_data - 1)
        offset_weights = np.insert(offset_weights, 0, 1)
    # normalize weights
    offset_weights = offset_weights / offset_weights.sum()
    # remove values from the upper 0.5 percentile
    high_mask = offset_weights.cumsum() > 0.995
    high_weights = offset_weights[high_mask].sum()
    offset_samples = offset_samples[~high_mask]
    offset_weights = offset_weights[~high_mask]
    offset_weights[-1] += high_weights
    # convert data to torch tensor
    offset_samples = torch.tensor(offset_samples, dtype=torch.int)
    offset_weights = torch.tensor(offset_weights)
    # thin offset histogram
    offset_samples, offset_weights = bin_hist(offset_samples, offset_weights, bin_size)

    data = defaultdict(lambda: None, data)
    target_xy = defaultdict(lambda: None, target_xy)
    # concatenate ontarget and offtarget
    is_ontarget = torch.cat(
        tuple(
            torch.full(
                target_xy[dtype].shape[:1], dtype == "ontarget", dtype=torch.bool
            )
            for dtype in glimpse.dtypes
        ),
        0,
    )
    data = torch.cat(tuple(data[dtype] for dtype in glimpse.dtypes), 0)
    target_xy = torch.cat(tuple(target_xy[dtype] for dtype in glimpse.dtypes), 0)
    time1 = torch.as_tensor(time1)
    ttb = torch.as_tensor(np.array(ttb)).T
    if all(labels[dtype] is None for dtype in glimpse.dtypes):
        labels = None
    else:
        labels = np.concatenate(
            tuple(
                labels[dtype] for dtype in glimpse.dtypes if labels[dtype] is not None
            ),
            0,
        )

    dataset = CosmosDataset(
        data,
        target_xy,
        is_ontarget,
        labels,
        offset_samples,
        offset_weights,
        time1=time1,
        ttb=ttb,
        name=name,
    )
    save(dataset, path)

    # plot offset distribution
    plt.figure(figsize=(3, 3))
    plt.bar(offset_samples, offset_weights, alpha=0.5, label="Offset")
    # plot data distribution for each channel
    for c in range(C):
        data_samples, data_counts = torch.unique(
            data[:, :, c], sorted=True, return_counts=True
        )
        data_weights = data_counts / data_counts.sum()
        plt.bar(data_samples, data_weights, alpha=0.5, label=f"Channel {c}")
    plt.title("Empirical Distribution", fontsize=12)
    plt.ylabel("Density", fontsize=12)
    plt.xlabel("Intensity", fontsize=12)
    plt.xlim(offset_samples.min(), dataset.vmax.max())
    plt.legend()
    plt.tight_layout()
    plt.savefig(path / "offset-distribution.png", dpi=300)

    plt.figure(figsize=(5, 3))
    plt.plot(offset_medians, label="Offset Median")
    plt.title("Offset drift", fontsize=12)
    plt.ylabel("Intensity", fontsize=12)
    plt.xlabel("Frames", fontsize=12)
    plt.ylim(offset_samples.min(), offset_samples.max())
    plt.legend()
    plt.tight_layout()
    plt.savefig(path / "offset-medians.png", dpi=300)

    logger.info(
        f"Dataset: N={dataset.N} on-target AOIs, "
        f"Nc={dataset.Nc} off-target AOIs, "
        f"F={dataset.F} frames, "
        f"C={dataset.C} channels, "
        f"Px={dataset.P} pixels, "
        f"Py={dataset.P} pixels"
    )
    logger.info(f"Data is saved in {Path(path) / 'data.tpqr'}")



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import funsor
from funsor.distribution import make_dist

from tapqir.distributions.affine_beta import AffineBeta
from tapqir.distributions.ksmogn import KSMOGN

funsor.set_backend("torch")

__all__ = [
    "AffineBeta",
    "KSMOGN",
]

FunsorAffineBeta = make_dist(AffineBeta)



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import pytest
import torch

from tapqir.utils.imscroll import (
    association_rate,
    count_intervals,
    dissociation_rate,
    time_to_first_binding,
)


@pytest.mark.parametrize(
    "labels,expected",
    [
        (
            np.array([[False, False, True], [True, False, True]]),
            np.array([[0, 2, -2], [0, 1, 3], [1, 1, -3], [1, 1, 0], [1, 1, 3]]),
        ),
        (
            np.array([[False, True, False], [True, True, False]]),
            np.array([[0, 1, -2], [0, 1, 1], [0, 1, 2], [1, 2, -3], [1, 1, 2]]),
        ),
        (
            torch.tensor([[False, False, True], [True, False, True]]),
            np.array([[0, 2, -2], [0, 1, 3], [1, 1, -3], [1, 1, 0], [1, 1, 3]]),
        ),
        (
            torch.tensor([[False, True, False], [True, True, False]]),
            np.array([[0, 1, -2], [0, 1, 1], [0, 1, 2], [1, 2, -3], [1, 1, 2]]),
        ),
    ],
)
def test_count_intervals(labels, expected):
    result = count_intervals(labels)
    actual = result[["aoi", "dwell_time", "low_or_high"]].values
    assert (actual == expected).all()


@pytest.mark.parametrize(
    "labels,expected",
    [
        (
            np.array(
                [
                    [False, False, False],
                    [False, False, True],
                    [False, True, True],
                    [True, False, True],
                ]
            ),
            np.array([3.0, 2.0, 1.0, 0.0]),
        ),
        (
            torch.tensor(
                [
                    [False, False, False],
                    [False, False, True],
                    [False, True, True],
                    [True, False, True],
                ]
            ),
            torch.tensor([3.0, 2.0, 1.0, 0.0]),
        ),
    ],
)
def test_time_to_first_binding(labels, expected):
    actual = time_to_first_binding(labels)
    assert (actual == expected).all()


@pytest.mark.parametrize(
    "labels,expected",
    [
        (
            np.array(
                [
                    [False, False, False, True, True],
                    [False, True, True, False, True],
                ]
            ),
            3 / 5,
        ),
        (
            np.array(
                [
                    [True, False, False, False, False],
                    [False, True, True, False, False],
                ]
            ),
            1 / 5,
        ),
    ],
)
def test_association_rate(labels, expected):
    actual = association_rate(labels)
    assert actual == expected


@pytest.mark.parametrize(
    "labels,expected",
    [
        (
            np.array(
                [
                    [False, False, False, True, True],
                    [False, True, True, False, True],
                ]
            ),
            1 / 3,
        ),
        (
            np.array(
                [
                    [True, False, False, False, False],
                    [False, True, True, False, False],
                ]
            ),
            2 / 3,
        ),
    ],
)
def test_dissociation_rate(labels, expected):
    actual = dissociation_rate(labels)
    assert actual == expected



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

from tapqir.models.cosmos import Cosmos
from tapqir.models.hmm import HMM
from tapqir.models.model import Model

__all__ = [
    "models",
    "Model",
    "Cosmos",
    "HMM",
]

models = {
    Cosmos.name: Cosmos,
    HMM.name: HMM,
}



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import math
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd
import pyro
import pyro.distributions as dist
import torch
from pyro.ops.indexing import Vindex
from pyro.ops.stats import hpdi, quantile
from sklearn.metrics import (
    confusion_matrix,
    matthews_corrcoef,
    precision_score,
    recall_score,
)

from tapqir.distributions import AffineBeta
from tapqir.distributions.util import gaussian_spots

pyro.enable_validation(False)


def snr_and_chi2(
    data: torch.Tensor,
    height: torch.Tensor,
    width: torch.Tensor,
    x: torch.Tensor,
    y: torch.Tensor,
    target_locs: torch.Tensor,
    background: torch.Tensor,
    gain: float,
    offset_mean: float,
    offset_var: float,
    P: int,
    theta_probs: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    r"""
    Calculate the signal-to-noise ratio.

    Total signal:

    .. math::
        \mu_{knf} =  \sum_{ij} I_{nfij}
        \mathcal{N}(i, j \mid x_{knf}, y_{knf}, w_{knf})`

    Noise:

    .. math::
        \sigma^2_{knf} = \sigma^2_{\text{offset}}
        + \mu_{knf} \text{gain}`

    Signal-to-noise ratio:

    .. math::
        \text{SNR}_{knf} =
        \dfrac{\mu_{knf} - b_{nf} - \mu_{\text{offset}}}{\sigma_{knf}}
        \text{ for } \theta_{nf} = k`
    """
    gaussians = gaussian_spots(
        height,
        width,
        x,
        y,
        target_locs,
        P,
    )

    # snr
    weights = gaussians / height[..., None, None]
    signal = ((data - background[..., None, None] - offset_mean) * weights).sum(
        dim=(-2, -1)
    )
    noise = (offset_var + background * gain).sqrt()
    snr_result = signal / noise
    mask = theta_probs > 0.5

    # chi2 test
    img_ideal = background[..., None, None] + gaussians.sum(-5)
    chi2_result = (data - img_ideal - offset_mean) ** 2 / img_ideal

    return snr_result[mask], chi2_result.mean(dim=(-1, -2))


def save_stats(model, path, CI=0.95, save_matlab=False):
    # global parameters
    global_params = model._global_params
    summary = pd.DataFrame(
        index=global_params,
        columns=["Mean", f"{int(100*CI)}% LL", f"{int(100*CI)}% UL"],
    )
    # local parameters
    local_params = [
        "height",
        "width",
        "x",
        "y",
        "background",
    ]

    ci_stats = defaultdict(partial(defaultdict, list))
    num_samples = 10000
    for param in global_params:
        if param == "gain":
            fn = dist.Gamma(
                pyro.param("gain_loc") * pyro.param("gain_beta"),
                pyro.param("gain_beta"),
            )
        elif param == "pi":
            fn = dist.Dirichlet(pyro.param("pi_mean") * pyro.param("pi_size"))
        elif param == "lamda":
            fn = dist.Gamma(
                pyro.param("lamda_loc") * pyro.param("lamda_beta"),
                pyro.param("lamda_beta"),
            )
        elif param == "proximity":
            fn = AffineBeta(
                pyro.param("proximity_loc"),
                pyro.param("proximity_size"),
                0,
                (model.data.P + 1) / math.sqrt(12),
            )
        elif param == "trans":
            fn = dist.Dirichlet(
                pyro.param("trans_mean") * pyro.param("trans_size")
            ).to_event(1)
        else:
            raise NotImplementedError
        samples = fn.sample((num_samples,)).data.squeeze()
        ci_stats[param] = {}
        LL, UL = hpdi(
            samples,
            CI,
            dim=0,
        )
        ci_stats[param]["LL"] = LL.cpu()
        ci_stats[param]["UL"] = UL.cpu()
        ci_stats[param]["Mean"] = fn.mean.data.squeeze().cpu()

        # calculate Keq
        if param == "pi":
            ci_stats["Keq"] = {}
            LL, UL = hpdi(samples[:, 1] / (1 - samples[:, 1]), CI, dim=0)
            ci_stats["Keq"]["LL"] = LL.cpu()
            ci_stats["Keq"]["UL"] = UL.cpu()
            ci_stats["Keq"]["Mean"] = (samples[:, 1] / (1 - samples[:, 1])).mean().cpu()

    # this does not need to be very accurate
    num_samples = 1000
    for param in local_params:
        LL, UL, Mean = [], [], []
        for ndx in torch.split(torch.arange(model.data.Nt), model.nbatch_size):
            ndx = ndx[:, None]
            kdx = torch.arange(model.K)[:, None, None]
            ll, ul, mean = [], [], []
            for fdx in torch.split(torch.arange(model.data.F), model.fbatch_size):
                if param == "background":
                    fn = dist.Gamma(
                        Vindex(pyro.param("b_loc"))[ndx, fdx]
                        * Vindex(pyro.param("b_beta"))[ndx, fdx],
                        Vindex(pyro.param("b_beta"))[ndx, fdx],
                    )
                elif param == "height":
                    fn = dist.Gamma(
                        Vindex(pyro.param("h_loc"))[kdx, ndx, fdx]
                        * Vindex(pyro.param("h_beta"))[kdx, ndx, fdx],
                        Vindex(pyro.param("h_beta"))[kdx, ndx, fdx],
                    )
                elif param == "width":
                    fn = AffineBeta(
                        Vindex(pyro.param("w_mean"))[kdx, ndx, fdx],
                        Vindex(pyro.param("w_size"))[kdx, ndx, fdx],
                        0.75,
                        2.25,
                    )
                elif param == "x":
                    fn = AffineBeta(
                        Vindex(pyro.param("x_mean"))[kdx, ndx, fdx],
                        Vindex(pyro.param("size"))[kdx, ndx, fdx],
                        -(model.data.P + 1) / 2,
                        (model.data.P + 1) / 2,
                    )
                elif param == "y":
                    fn = AffineBeta(
                        Vindex(pyro.param("y_mean"))[kdx, ndx, fdx],
                        Vindex(pyro.param("size"))[kdx, ndx, fdx],
                        -(model.data.P + 1) / 2,
                        (model.data.P + 1) / 2,
                    )
                else:
                    raise NotImplementedError
                samples = fn.sample((num_samples,)).data
                l, u = hpdi(
                    samples,
                    CI,
                    dim=0,
                )
                m = fn.mean.data
                ll.append(l)
                ul.append(u)
                mean.append(m)
            else:
                LL.append(torch.cat(ll, -1))
                UL.append(torch.cat(ul, -1))
                Mean.append(torch.cat(mean, -1))
        else:
            ci_stats[param]["LL"] = torch.cat(LL, -2).cpu()
            ci_stats[param]["UL"] = torch.cat(UL, -2).cpu()
            ci_stats[param]["Mean"] = torch.cat(Mean, -2).cpu()

    for param in global_params:
        if param == "pi":
            summary.loc[param, "Mean"] = ci_stats[param]["Mean"][1].item()
            summary.loc[param, "95% LL"] = ci_stats[param]["LL"][1].item()
            summary.loc[param, "95% UL"] = ci_stats[param]["UL"][1].item()
            # Keq
            summary.loc["Keq", "Mean"] = ci_stats["Keq"]["Mean"].item()
            summary.loc["Keq", "95% LL"] = ci_stats["Keq"]["LL"].item()
            summary.loc["Keq", "95% UL"] = ci_stats["Keq"]["UL"].item()
        elif param == "trans":
            summary.loc["kon", "Mean"] = ci_stats[param]["Mean"][0, 1].item()
            summary.loc["kon", "95% LL"] = ci_stats[param]["LL"][0, 1].item()
            summary.loc["kon", "95% UL"] = ci_stats[param]["UL"][0, 1].item()
            summary.loc["koff", "Mean"] = ci_stats[param]["Mean"][1, 0].item()
            summary.loc["koff", "95% LL"] = ci_stats[param]["LL"][1, 0].item()
            summary.loc["koff", "95% UL"] = ci_stats[param]["UL"][1, 0].item()
        else:
            summary.loc[param, "Mean"] = ci_stats[param]["Mean"].item()
            summary.loc[param, "95% LL"] = ci_stats[param]["LL"].item()
            summary.loc[param, "95% UL"] = ci_stats[param]["UL"].item()
    ci_stats["m_probs"] = model.m_probs.data.cpu()
    ci_stats["theta_probs"] = model.theta_probs.data.cpu()
    ci_stats["z_probs"] = model.z_probs.data.cpu()
    ci_stats["z_map"] = model.z_map.data.cpu()
    ci_stats["p_specific"] = ci_stats["theta_probs"].sum(0)

    # calculate vmin/vmax
    theta_mask = ci_stats["theta_probs"] > 0.5
    if theta_mask.sum():
        hmax = quantile(ci_stats["height"]["Mean"][theta_mask], 0.99)
    else:
        hmax = 1
    ci_stats["height"]["vmin"] = -0.03 * hmax
    ci_stats["height"]["vmax"] = 1.3 * hmax
    ci_stats["width"]["vmin"] = 0.5
    ci_stats["width"]["vmax"] = 2.5
    ci_stats["x"]["vmin"] = -9
    ci_stats["x"]["vmax"] = 9
    ci_stats["y"]["vmin"] = -9
    ci_stats["y"]["vmax"] = 9
    bmax = quantile(ci_stats["background"]["Mean"].flatten(), 0.99)
    ci_stats["background"]["vmin"] = -0.03 * bmax
    ci_stats["background"]["vmax"] = 1.3 * bmax

    # timestamps
    if model.data.time1 is not None:
        ci_stats["time1"] = model.data.time1
    if model.data.ttb is not None:
        ci_stats["ttb"] = model.data.ttb

    # intensity of target-specific spots
    theta_mask = torch.argmax(ci_stats["theta_probs"], dim=0)
    h_specific = Vindex(ci_stats["height"]["Mean"])[
        theta_mask, torch.arange(model.data.Nt)[:, None], torch.arange(model.data.F)
    ]
    ci_stats["h_specific"] = h_specific * (ci_stats["z_map"] > 0).long()

    model.params = ci_stats

    # snr and chi2 test
    snr_result, chi2_result = snr_and_chi2(
        model.data.images[:, :, model.cdx],
        ci_stats["height"]["Mean"],
        ci_stats["width"]["Mean"],
        ci_stats["x"]["Mean"],
        ci_stats["y"]["Mean"],
        model.data.xy[:, :, model.cdx],
        ci_stats["background"]["Mean"],
        ci_stats["gain"]["Mean"],
        model.data.offset.mean,
        model.data.offset.var,
        model.data.P,
        model.theta_probs,
    )
    summary.loc["SNR", "Mean"] = snr_result.mean().item()
    ci_stats["chi2"] = {}
    ci_stats["chi2"]["values"] = chi2_result
    cmax = quantile(ci_stats["chi2"]["values"].flatten(), 0.99)
    ci_stats["chi2"]["vmin"] = -0.03 * cmax
    ci_stats["chi2"]["vmax"] = 1.3 * cmax

    # classification statistics
    if model.data.labels is not None:
        pred_labels = model.z_map[model.data.is_ontarget].cpu().numpy().ravel()
        true_labels = model.data.labels["z"][: model.data.N, :, model.cdx].ravel()

        with np.errstate(divide="ignore", invalid="ignore"):
            summary.loc["MCC", "Mean"] = matthews_corrcoef(true_labels, pred_labels)
        summary.loc["Recall", "Mean"] = recall_score(
            true_labels, pred_labels, zero_division=0
        )
        summary.loc["Precision", "Mean"] = precision_score(
            true_labels, pred_labels, zero_division=0
        )

        (
            summary.loc["TN", "Mean"],
            summary.loc["FP", "Mean"],
            summary.loc["FN", "Mean"],
            summary.loc["TP", "Mean"],
        ) = confusion_matrix(true_labels, pred_labels, labels=(0, 1)).ravel()

        mask = torch.from_numpy(model.data.labels["z"][: model.data.N, :, model.cdx])
        samples = torch.masked_select(model.z_probs[model.data.is_ontarget].cpu(), mask)
        if len(samples):
            z_ll, z_ul = hpdi(samples, CI)
            summary.loc["p(specific)", "Mean"] = quantile(samples, 0.5).item()
            summary.loc["p(specific)", "95% LL"] = z_ll.item()
            summary.loc["p(specific)", "95% UL"] = z_ul.item()
        else:
            summary.loc["p(specific)", "Mean"] = 0.0
            summary.loc["p(specific)", "95% LL"] = 0.0
            summary.loc["p(specific)", "95% UL"] = 0.0

    model.summary = summary

    if path is not None:
        path = Path(path)
        torch.save(ci_stats, path / f"{model.full_name}-params.tpqr")
        if save_matlab:
            from scipy.io import savemat

            for param, field in ci_stats.items():
                if param in (
                    "m_probs",
                    "theta_probs",
                    "z_probs",
                    "z_map",
                    "p_specific",
                    "h_specific",
                    "time1",
                    "ttb",
                ):
                    ci_stats[param] = field.numpy()
                    continue
                for stat, value in field.items():
                    ci_stats[param][stat] = value.cpu().numpy()
            savemat(path / f"{model.full_name}-params.mat", ci_stats)
        summary.to_csv(
            path / f"{model.full_name}-summary.csv",
        )



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

from tapqir.imscroll.glimpse_reader import GlimpseDataset, read_glimpse

__all__ = [
    "GlimpseDataset",
    "read_glimpse",
]



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging
import sys
from collections import defaultdict
from enum import Enum
from functools import partial
from pathlib import Path
from typing import List, Optional

import colorama
import matplotlib.pyplot as plt
import torch
import typer
import yaml
from tqdm import tqdm

from tapqir import __version__
from tapqir.distributions.util import gaussian_spots

app = typer.Typer()

# option DEFAULTS
DEFAULTS = defaultdict(lambda: None)
DEFAULTS["channels"] = []


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


def format_link(link):
    return "<{blue}{link}{nc}>".format(
        blue=colorama.Fore.CYAN, link=link, nc=colorama.Fore.RESET
    )


# available models
class Model(str, Enum):
    cosmos = "cosmos"


def get_default(key):
    if key in [
        "name",
        "glimpse-folder",
        "ontarget-aoiinfo",
        "offtarget-aoiinfo",
        "driftlist",
    ]:
        return [c.get(key, None) for c in DEFAULTS["channels"]]
    return DEFAULTS[key]


def deactivate_prompts(ctx, param, value):
    if param.name == "no_input" and value:
        for p in ctx.command.params:
            if isinstance(p, typer.core.TyperOption) and p.prompt is not None:
                p.prompt = None
    elif param.name == "frame_range" and not value:
        for p in ctx.command.params:
            if p.name in ("frame_start", "frame_end"):
                p.prompt = None
    return value


@app.command()
def glimpse(
    dataset: str = typer.Option(
        partial(get_default, "dataset"), help="Dataset name", prompt="Dataset name"
    ),
    P: int = typer.Option(
        partial(get_default, "P"),
        "--aoi-size",
        "-P",
        help="AOI image size - number of pixels along the axis",
        prompt="AOI image size - number of pixels along the axis",
    ),
    frame_range: bool = typer.Option(
        partial(get_default, "frame-range"),
        help="Specify frame range.",
        prompt="Specify frame range?",
        callback=deactivate_prompts,
    ),
    frame_start: Optional[int] = typer.Option(
        partial(get_default, "frame-start"),
        help="Starting frame.",
        prompt="Starting frame",
    ),
    frame_end: Optional[int] = typer.Option(
        partial(get_default, "frame-end"),
        help="Ending frame.",
        prompt="Ending frame",
    ),
    num_channels: int = typer.Option(
        partial(get_default, "num-channels"),
        "--num-channels",
        "-C",
        help="Number of color channels",
        prompt="Number of color channels",
    ),
    use_offtarget: bool = typer.Option(
        partial(get_default, "use-offtarget"),
        help="Use off-target AOI locations.",
        prompt="Use off-target AOI locations?",
    ),
    name: Optional[List[str]] = typer.Option(partial(get_default, "name")),
    glimpse_folder: Optional[List[str]] = typer.Option(
        partial(get_default, "glimpse-folder")
    ),
    ontarget_aoiinfo: Optional[List[str]] = typer.Option(
        partial(get_default, "ontarget-aoiinfo")
    ),
    offtarget_aoiinfo: Optional[List[str]] = typer.Option(
        partial(get_default, "offtarget-aoiinfo")
    ),
    driftlist: Optional[List[str]] = typer.Option(partial(get_default, "driftlist")),
    overwrite: bool = typer.Option(
        True,
        "--overwrite",
        "-w",
        help="Overwrite defaults values.",
        prompt="Overwrite defaults values?",
    ),
    no_input: bool = typer.Option(
        False,
        "--no-input",
        help="Disable interactive prompt.",
        is_eager=True,
        callback=deactivate_prompts,
    ),
    labels: bool = typer.Option(
        False,
        "--labels",
        "-l",
        help="Add on-target binding labels.",
    ),
    progress_bar=None,
):
    """
    Extract AOIs from raw glimpse images.

    Analyzing data acquired with Glimpse and pre-processed with  the imscroll program
    will require the following files:

    * image data in glimpse format and header file\n
    * aoiinfo file designating the locations of target molecules (on-target AOIs) in the binder channel\n
    * (optional) aoiinfo file designating the off-target control locations (off-target AOIs) in the binder channel\n
    * driftlist file recording the stage movement that took place during the experiment
    """

    from tapqir.imscroll import read_glimpse

    torch.set_default_tensor_type(torch.FloatTensor)

    global DEFAULTS
    cd = DEFAULTS["cd"]

    if progress_bar is None:
        progress_bar = tqdm

    # fill in default values
    DEFAULTS["dataset"] = dataset
    DEFAULTS["P"] = P
    DEFAULTS["num-channels"] = num_channels
    DEFAULTS["frame-range"] = frame_range
    DEFAULTS["frame-start"] = frame_start
    DEFAULTS["frame-end"] = frame_end
    DEFAULTS["use-offtarget"] = use_offtarget
    DEFAULTS["labels"] = labels

    # inputs descriptions
    desc = {}
    desc["name"] = "Channel name"
    desc["glimpse-folder"] = "Header/glimpse folder"
    desc["driftlist"] = "Driftlist file"
    desc["ontarget-aoiinfo"] = "Target molecule locations file"
    desc["offtarget-aoiinfo"] = "Off-target control locations file"
    desc["ontarget-labels"] = "On-target AOI binding labels"
    desc["offtarget-labels"] = "Off-target AOI binding labels"

    keys = [
        "name",
        "glimpse-folder",
        "ontarget-aoiinfo",
        "offtarget-aoiinfo",
        "driftlist",
    ]
    flags = [name, glimpse_folder, ontarget_aoiinfo, offtarget_aoiinfo, driftlist]

    for c in range(num_channels):
        if len(DEFAULTS["channels"]) < c + 1:
            DEFAULTS["channels"].append({})
        for flag, key in zip(flags, keys):
            DEFAULTS["channels"][c][key] = flag[c] if c < len(flag) else None

    # interactive input
    if not no_input:
        for c in range(num_channels):
            if labels:
                keys += ["ontarget-labels", "offtarget-labels"]
            typer.echo(f"\nINPUTS FOR CHANNEL #{c}\n")
            for key in keys:
                if key == "offtarget-aoiinfo" and not use_offtarget:
                    continue

                DEFAULTS["channels"][c][key] = typer.prompt(
                    desc[key], default=DEFAULTS["channels"][c][key]
                )

    DEFAULTS = dict(DEFAULTS)
    for c in range(num_channels):
        DEFAULTS["channels"][c] = dict(DEFAULTS["channels"][c])

    if overwrite:
        with open(cd / ".tapqir" / "config.yaml", "w") as cfg_file:
            yaml.dump(
                {key: value for key, value in DEFAULTS.items() if key != "cd"},
                cfg_file,
                sort_keys=False,
            )

    typer.echo("Extracting AOIs ...")
    read_glimpse(
        path=cd,
        progress_bar=progress_bar,
        **DEFAULTS,
    )
    typer.echo("Extracting AOIs: Done")


@app.command()
def fit(
    model: Model = typer.Option("cosmos", help="Tapqir model", prompt="Tapqir model"),
    channels: List[int] = typer.Option(
        [0],
        help="Color-channel numbers to analyze",
        prompt="Channel numbers (space separated if multiple)",
    ),
    cuda: bool = typer.Option(
        partial(get_default, "cuda"),
        "--cuda/--cpu",
        help="Run computations on GPU or CPU",
        prompt="Run computations on GPU?",
        show_default=False,
    ),
    nbatch_size: int = typer.Option(
        partial(get_default, "nbatch-size"),
        "--nbatch-size",
        "-n",
        help="AOI batch size",
        prompt="AOI batch size",
    ),
    fbatch_size: int = typer.Option(
        partial(get_default, "fbatch-size"),
        "--fbatch-size",
        "-f",
        help="Frame batch size",
        prompt="Frame batch size",
    ),
    learning_rate: float = typer.Option(
        partial(get_default, "learning-rate"),
        "--learning-rate",
        "-lr",
        help="Learning rate",
        prompt="Learning rate",
    ),
    num_iter: int = typer.Option(
        0,
        "--num-iter",
        "-it",
        help="Number of iterations",
        prompt="Number of iterations",
    ),
    k_max: int = typer.Option(
        2, "--k-max", "-k", help="Maximum number of spots per image"
    ),
    matlab: bool = typer.Option(
        partial(get_default, "matlab"),
        "--matlab",
        help="Save parameters in matlab format",
        prompt="Save parameters in matlab format?",
    ),
    funsor: bool = typer.Option(
        False, "--funsor/--pyro", help="Use funsor or pyro backend"
    ),
    pykeops: bool = typer.Option(
        True,
        "--pykeops/--no-pykeops",
        help="Use pykeops backend for offset marginalization",
    ),
    overwrite: bool = typer.Option(
        True,
        "--overwrite",
        "-w",
        help="Overwrite defaults values.",
        prompt="Overwrite defaults values?",
    ),
    no_input: bool = typer.Option(
        False,
        "--no-input",
        help="Disable interactive prompt.",
        is_eager=True,
        callback=deactivate_prompts,
    ),
    progress_bar=None,
):
    """
    Fit the data to the selected model.

    Available models:

    * cosmos: single-color time-independent co-localization model.\n
    """
    global DEFAULTS
    cd = DEFAULTS["cd"]

    if progress_bar is None:
        progress_bar = tqdm

    from pyroapi import pyro_backend

    from tapqir.models import models
    from tapqir.utils.stats import save_stats

    settings = {}
    settings["S"] = 1
    settings["K"] = k_max
    settings["channels"] = channels
    settings["device"] = "cuda" if cuda else "cpu"
    settings["dtype"] = "double"
    settings["use_pykeops"] = pykeops
    # priors settings
    settings["background_mean_std"] = DEFAULTS["background_mean_std"]
    settings["background_std_std"] = DEFAULTS["background_std_std"]
    settings["lamda_rate"] = DEFAULTS["lamda_rate"]
    settings["height_std"] = DEFAULTS["height_std"]
    settings["width_min"] = DEFAULTS["width_min"]
    settings["width_max"] = DEFAULTS["width_max"]
    settings["proximity_rate"] = DEFAULTS["proximity_rate"]
    settings["gain_std"] = DEFAULTS["gain_std"]

    if overwrite:
        DEFAULTS["cuda"] = cuda
        DEFAULTS["nbatch-size"] = nbatch_size
        DEFAULTS["fbatch-size"] = fbatch_size
        DEFAULTS["learning-rate"] = learning_rate
        DEFAULTS["matlab"] = matlab
        with open(cd / ".tapqir" / "config.yaml", "w") as cfg_file:
            yaml.dump(
                {key: value for key, value in DEFAULTS.items() if key != "cd"},
                cfg_file,
                sort_keys=False,
            )

    backend = "funsor" if funsor else "pyro"
    if backend == "pyro":
        PYRO_BACKEND = "pyro"
    elif backend == "funsor":
        import funsor
        import pyro.contrib.funsor  # noqa: F401

        funsor.set_backend("torch")
        PYRO_BACKEND = "contrib.funsor"
    else:
        raise ValueError("Only pyro and funsor backends are supported.")

    with pyro_backend(PYRO_BACKEND):

        model = models[model](**settings)
        model.load(cd)

        model.init(learning_rate, nbatch_size, fbatch_size)
        typer.echo("Fitting the data ...")
        exit_code = model.run(num_iter, progress_bar=progress_bar)
        if exit_code == 0:
            typer.echo("Fitting the data: Done")
        elif exit_code == 1:
            typer.echo("The model hasn't converged!")
        typer.echo("Computing stats ...")
        save_stats(model, cd, save_matlab=matlab)
        typer.echo("Computing stats: Done")


@app.command()
def stats(
    model: Model = typer.Option("cosmos", help="Tapqir model", prompt="Tapqir model"),
    channels: List[int] = typer.Option(
        [0],
        help="Color-channel numbers to analyze",
        prompt="Channel numbers (space separated if multiple)",
    ),
    cuda: bool = typer.Option(
        partial(get_default, "cuda"),
        "--cuda/--cpu",
        help="Run computations on GPU or CPU",
        prompt="Run computations on GPU?",
        show_default=False,
    ),
    nbatch_size: int = typer.Option(
        partial(get_default, "nbatch-size"),
        "--nbatch-size",
        "-n",
        help="AOI batch size",
        prompt="AOI batch size",
    ),
    fbatch_size: int = typer.Option(
        partial(get_default, "fbatch-size"),
        "--fbatch-size",
        "-f",
        help="Frame batch size",
        prompt="Frame batch size",
    ),
    matlab: bool = typer.Option(
        partial(get_default, "matlab"),
        "--matlab",
        help="Save parameters in matlab format",
        prompt="Save parameters in matlab format?",
    ),
    funsor: bool = typer.Option(
        False, "--funsor/--pyro", help="Use funsor or pyro backend"
    ),
    no_input: bool = typer.Option(
        False,
        "--no-input",
        help="Use defaults values.",
        is_eager=True,
        callback=deactivate_prompts,
    ),
):
    from pyroapi import pyro_backend

    from tapqir.models import models
    from tapqir.utils.stats import save_stats

    global DEFAULTS
    cd = DEFAULTS["cd"]

    dtype = "double"
    device = "cuda" if cuda else "cpu"
    backend = "funsor" if funsor else "pyro"

    settings = {}
    settings["S"] = 1
    settings["channels"] = channels
    settings["device"] = device
    settings["dtype"] = dtype

    # pyro backend
    if backend == "pyro":
        PYRO_BACKEND = "pyro"
    elif backend == "funsor":
        import funsor
        import pyro.contrib.funsor  # noqa: F401

        funsor.set_backend("torch")
        PYRO_BACKEND = "contrib.funsor"
    else:
        raise ValueError("Only pyro and funsor backends are supported.")

    with pyro_backend(PYRO_BACKEND):

        model = models[model](**settings)
        model.load(cd)
        model.load_checkpoint(param_only=True)
        model.nbatch_size = nbatch_size
        model.fbatch_size = fbatch_size

        typer.echo("Computing stats ...")
        save_stats(model, cd, save_matlab=matlab)
        typer.echo("Computing stats: Done")


def config_axis(ax, label, f1, f2, ymin=None, ymax=None, xticklabels=False):
    plt.minorticks_on()
    ax.tick_params(
        direction="in",
        which="minor",
        length=1,
        bottom=True,
        top=True,
        left=True,
        right=True,
    )
    ax.tick_params(
        direction="in",
        which="major",
        length=2,
        bottom=True,
        top=True,
        left=True,
        right=True,
    )
    ax.set_ylabel(label)
    ax.set_xlim(f1 - 0.5, f2 - 0.5)
    if (ymin is not None) and (ymax is not None):
        ax.set_ylim(ymin, ymax)
    if not xticklabels:
        ax.set_xticklabels([])


@app.command()
def show(
    model: Model = typer.Option("cosmos", help="Tapqir model", prompt="Tapqir model"),
    channels: List[int] = typer.Option(
        [0],
        help="Color-channel numbers to analyze",
        prompt="Channel numbers (space separated if multiple)",
    ),
    n: int = typer.Option(0, help="n", prompt="n"),
    f1: Optional[int] = None,
    f2: Optional[int] = None,
):
    from tapqir.models import models

    global DEFAULTS
    cd = DEFAULTS["cd"]

    model = models[model](1, 2, channels, "cpu", "float")
    model.load(cd, data_only=False)
    if f1 is None:
        f1 = 0
    if f2 is None:
        f2 = f1 + 15
    c = model.cdx

    width, height, dpi = 6.25, 6.25, 100
    fig = plt.figure(figsize=(width, height), dpi=dpi)
    gs = fig.add_gridspec(
        nrows=10,
        ncols=15,
        top=0.96,
        bottom=0.04,
        left=0.1,
        right=0.98,
        hspace=0.1,
        height_ratios=[0.9, 0.9, 1, 1, 1, 1, 1, 1, 1, 1],
    )
    ax = {}
    item = {}

    frames = torch.arange(f1, f2)
    img_ideal = (
        model.data.offset.mean
        + model.params["background"]["Mean"][n, frames, None, None]
    )
    gaussian = gaussian_spots(
        model.params["height"]["Mean"][:, n, frames],
        model.params["width"]["Mean"][:, n, frames],
        model.params["x"]["Mean"][:, n, frames],
        model.params["y"]["Mean"][:, n, frames],
        model.data.xy[n, frames, c],
        model.data.P,
    )
    img_ideal = img_ideal + gaussian.sum(-4)
    for f in range(15):
        ax[f"image_{f}"] = fig.add_subplot(gs[0, f])
        item[f"image_{f}"] = ax[f"image_{f}"].imshow(
            model.data.images[n, f, c].numpy(),
            vmin=model.data.vmin[model.cdx] - 50,
            vmax=model.data.vmax[model.cdx] + 50,
            cmap="gray",
        )
        ax[f"image_{f}"].set_title(rf"${f}$", fontsize=9)
        ax[f"image_{f}"].axis("off")
        ax[f"ideal_{f}"] = fig.add_subplot(gs[1, f])
        item[f"ideal_{f}"] = ax[f"ideal_{f}"].imshow(
            img_ideal[f].numpy(),
            vmin=model.data.vmin[model.cdx] - 50,
            vmax=model.data.vmax[model.cdx] + 50,
            cmap="gray",
        )
        ax[f"ideal_{f}"].axis("off")

    ax["z_map"] = fig.add_subplot(gs[2, :])
    config_axis(ax["z_map"], r"$z$", f1, f2, -0.1, model.S + 0.1)
    (item["z_map"],) = ax["z_map"].plot(
        torch.arange(0, model.data.F),
        model.params["z_map"][n],
        "-",
        lw=1,
        color="k",
    )

    ax["p_specific"] = fig.add_subplot(gs[3, :])
    config_axis(ax["p_specific"], r"$p(\mathsf{specific})$", f1, f2, -0.1, 1.1)
    (item["p_specific"],) = ax["p_specific"].plot(
        torch.arange(0, model.data.F),
        model.params["p_specific"][n],
        "o-",
        ms=3,
        lw=1,
        color="C2",
    )

    ax["height"] = fig.add_subplot(gs[4, :])
    config_axis(
        ax["height"],
        r"$h$",
        f1,
        f2,
        model.params["height"]["vmin"],
        model.params["height"]["vmax"],
    )

    ax["width"] = fig.add_subplot(gs[5, :])
    config_axis(
        ax["width"],
        r"$w$",
        f1,
        f2,
        model.params["width"]["vmin"],
        model.params["width"]["vmax"],
    )

    ax["x"] = fig.add_subplot(gs[6, :])
    config_axis(
        ax["x"], r"$x$", f1, f2, model.params["x"]["vmin"], model.params["x"]["vmax"]
    )

    ax["y"] = fig.add_subplot(gs[7, :])
    config_axis(
        ax["y"], r"$y$", f1, f2, model.params["y"]["vmin"], model.params["y"]["vmax"]
    )

    ax["background"] = fig.add_subplot(gs[8, :])
    config_axis(
        ax["background"],
        r"$b$",
        f1,
        f2,
        model.params["background"]["vmin"],
        model.params["background"]["vmax"],
    )

    ax["chi2"] = fig.add_subplot(gs[9, :])
    config_axis(
        ax["chi2"],
        r"$\chi^2 \mathsf{test}$",
        f1,
        f2,
        model.params["chi2"]["vmin"],
        model.params["chi2"]["vmax"],
        True,
    )
    ax["chi2"].set_xlabel("Time (frame)")

    theta_mask = model.params["theta_probs"][:, n] > 0.5
    j_mask = (model.params["m_probs"][:, n] > 0.5) & ~theta_mask
    for p in ["height", "width", "x", "y"]:
        # target-nonspecific spots
        for k in range(model.K):
            f_mask = j_mask[k]
            mean = model.params[p]["Mean"][k, n] * f_mask
            ll = model.params[p]["LL"][k, n] * f_mask
            ul = model.params[p]["UL"][k, n] * f_mask
            (item[f"{p}_nonspecific{k}_mean"],) = ax[p].plot(
                torch.arange(0, model.data.F)[f_mask],
                mean[f_mask],
                "o",
                ms=2,
                lw=1,
                color=f"C{k}",
            )
            item[f"{p}_nonspecific{k}_fill"] = ax[p].fill_between(
                torch.arange(0, model.data.F),
                ll,
                ul,
                where=f_mask,
                alpha=0.3,
                color=f"C{k}",
            )
        # target-specific spots
        f_mask = theta_mask.sum(0).bool()
        mean = (model.params[p]["Mean"][:, n] * theta_mask).sum(0)
        ll = (model.params[p]["LL"][:, n] * theta_mask).sum(0)
        ul = (model.params[p]["UL"][:, n] * theta_mask).sum(0)
        (item[f"{p}_specific_mean"],) = ax[p].plot(
            torch.arange(0, model.data.F)
            if p == "height"
            else torch.arange(0, model.data.F)[f_mask],
            mean if p == "height" else mean[f_mask],
            "-" if p == "height" else "o",
            ms=2,
            color="C2",
        )
        item[f"{p}_specific_fill"] = ax[p].fill_between(
            torch.arange(0, model.data.F),
            ll,
            ul,
            where=f_mask,
            alpha=0.3,
            color="C2",
        )
    (item["background_mean"],) = ax["background"].plot(
        torch.arange(0, model.data.F),
        model.params["background"]["Mean"][n],
        "o-",
        ms=3,
        lw=1,
        color="k",
    )
    item["background_fill"] = ax["background"].fill_between(
        torch.arange(0, model.data.F),
        model.params["background"]["LL"][n],
        model.params["background"]["UL"][n],
        alpha=0.3,
        color="k",
    )

    (item["chi2"],) = ax["chi2"].plot(
        torch.arange(0, model.data.F),
        model.params["chi2"]["values"][n],
        "-",
        lw=1,
        color="k",
    )

    plt.show()
    return model, fig, item, ax


@app.command()
def log():
    """Show logging info."""
    import pydoc

    global DEFAULTS
    cd = DEFAULTS["cd"]

    log_file = cd / ".tapqir" / "loginfo"
    with open(log_file, "r") as f:
        pydoc.pager(f.read())


@app.callback()
def main(
    cd: Path = typer.Option(
        Path.cwd(),
        help="Change working directory.",
        show_default=False,
        exists=True,
        file_okay=False,
        dir_okay=True,
    ),
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        callback=version_callback,
        help="Show Tapqir version and exit.",
        is_eager=True,
    ),
):
    """
    Bayesian analysis of co-localization single-molecule microscopy image data.

    Initialize Tapqir in the working directory. Initializing a Tapqir workspace creates
    a ``.tapqir`` sub-directory for storing ``config.yaml`` file, ``loginfo`` file,
    and files that are created by commands such as ``tapqir fit``.
    """

    global DEFAULTS
    # set working directory
    DEFAULTS["cd"] = cd

    TAPQIR_PATH = cd / ".tapqir"

    if not TAPQIR_PATH.is_dir():
        # initialize directory
        TAPQIR_PATH.mkdir()
    CONFIG_PATH = TAPQIR_PATH / "config.yaml"
    if not CONFIG_PATH.is_file():
        with open(CONFIG_PATH, "w") as cfg_file:
            # model settings
            DEFAULTS["P"] = 14
            DEFAULTS["nbatch-size"] = 10
            DEFAULTS["fbatch-size"] = 512
            DEFAULTS["learning-rate"] = 0.005
            DEFAULTS["num-channels"] = 1
            DEFAULTS["cuda"] = True
            DEFAULTS["matlab"] = False
            # priors settings
            DEFAULTS["background_mean_std"] = 1000
            DEFAULTS["background_std_std"] = 100
            DEFAULTS["lamda_rate"] = 1
            DEFAULTS["height_std"] = 10000
            DEFAULTS["width_min"] = 0.75
            DEFAULTS["width_max"] = 2.25
            DEFAULTS["proximity_rate"] = 1
            DEFAULTS["gain_std"] = 50
            # offset settings
            DEFAULTS["offset_x"] = 10
            DEFAULTS["offset_y"] = 10
            DEFAULTS["offset_P"] = 30
            DEFAULTS["bin_size"] = 1
            # save config file
            yaml.dump(
                {key: value for key, value in DEFAULTS.items() if key != "cd"},
                cfg_file,
                sort_keys=False,
            )

        typer.echo(
            (
                f"Initialized Tapqir at {TAPQIR_PATH}.\n"
                "{yellow}---------------------------------------------------------------{nc}\n"
                f"- Checkout the documentation: {format_link('https://tapqir.readthedocs.io/')}\n"
                f"- Get help on our forum: {format_link('https://github.com/gelles-brandeis/tapqir/discussions')}\n"
            ).format(yellow=colorama.Fore.YELLOW, nc=colorama.Fore.RESET)
        )

    # read defaults from config file
    with open(CONFIG_PATH, "r") as cfg_file:
        cfg_defaults = yaml.safe_load(cfg_file) or {}
        DEFAULTS.update(cfg_defaults)
        typer.echo(f"Configuration options are read from {CONFIG_PATH}.")

    # create logger
    logger = logging.getLogger("tapqir")

    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter(
        fmt="%(levelname)s - %(message)s",
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    fh = logging.FileHandler(cd / ".tapqir" / "loginfo")
    fh.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        fmt="%(asctime)s - %(levelname)s - %(message)s",
        datefmt="%m/%d/%Y %I:%M %p",
    )
    fh.setFormatter(formatter)
    logger.addHandler(fh)


# click object is required to generate docs with sphinx-click
typer_click_object = typer.main.get_command(app)

if __name__ == "__main__":
    app()



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

"""
Distribution utility functions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import math
from functools import lru_cache

import torch


def gaussian_spots(
    height: torch.Tensor,
    width: torch.Tensor,
    x: torch.Tensor,
    y: torch.Tensor,
    target_locs: torch.Tensor,
    P: int,
    m: torch.Tensor = None,
) -> torch.Tensor:
    r"""
    Calculates ideal shape of the 2D-Gaussian spots given spot parameters
    and target positions.

    .. math::
        \mu^S_{\mathsf{pixelX}(i), \mathsf{pixelY}(j)} =
        \dfrac{m \cdot h}{2 \pi w^2}
        \exp{\left( -\dfrac{(i-x-x^\mathsf{target})^2 + (j-y-y^\mathsf{target})^2}{2 w^2} \right)}

    :param height: Integrated spot intensity. Should be broadcastable to ``batch_shape``.
    :param width: Spot width. Should be broadcastable to ``batch_shape``.
    :param x: Spot center on x-axis. Should be broadcastable to ``batch_shape``.
    :param y: Spot center on y-axis. Should be broadcastable to ``batch_shape``.
    :param target_locs: Target location. Should have
        the rightmost size ``2`` correspondnig to locations on
        x- and y-axes, and be broadcastable to ``batch_shape + (2,)``.
    :param P: Number of pixels along the axis.
    :param m: Spot presence indicator. Should be broadcastable to ``batch_shape``.
    :return: A tensor of a shape ``batch_shape + (P, P)`` representing 2D-Gaussian spots.
    """
    # create meshgrid of PxP pixel positions
    device = height.device
    P_range = torch.arange(P, device=device)
    i_pixel, j_pixel = torch.meshgrid(P_range, P_range, indexing="xy")
    ij_pixel = torch.stack((i_pixel, j_pixel), dim=-1)

    # Ideal 2D gaussian spots
    spot_locs = target_locs + torch.stack((x, y), -1)
    scale = width[..., None, None, None]
    loc = spot_locs[..., None, None, :]
    var = scale**2
    normalized_gaussian = torch.exp(
        (
            -((ij_pixel - loc) ** 2) / (2 * var)
            - scale.log()
            - math.log(math.sqrt(2 * math.pi))
        ).sum(-1)
    )
    if m is not None:
        height = m * height
    return height[..., None, None] * normalized_gaussian


def truncated_poisson_probs(lamda: torch.Tensor, K: int) -> torch.Tensor:
    r"""
    Probability of the number of non-specific spots.

    .. math::

        \mathbf{TruncatedPoisson}(\lambda, K) =
        \begin{cases}
            1 - e^{-\lambda} \sum_{i=0}^{K-1} \dfrac{\lambda^i}{i!} & \textrm{if } k = K \\
            \dfrac{\lambda^k e^{-\lambda}}{k!} & \mathrm{otherwise}
        \end{cases}

    :param lamda: Average rate of target-nonspecific binding.
    :param K: Maximum number of spots that can be present in a single image.
    :return: A tensor of a shape ``lamda.shape + (K+1,)`` of probabilities.
    """
    shape = lamda.shape + (K + 1,)
    dtype = lamda.dtype
    result = torch.zeros(shape, dtype=dtype)
    kdx = torch.arange(K)
    result[..., :-1] = torch.exp(
        kdx.xlogy(lamda.unsqueeze(-1)) - lamda.unsqueeze(-1) - (kdx + 1).lgamma()
    )
    result[..., -1] = 1 - result[..., :-1].sum(-1)
    return result


def probs_m(lamda: torch.Tensor, K: int) -> torch.Tensor:
    r"""
    Prior spot presence probability :math:`p(m | \theta, \lambda)`.

    .. math::

        p(m_{\mathsf{spot}(k)} | \theta, \lambda) =
        \begin{cases}
            \mathbf{Bernoulli}(1) & \text{$\theta = k$} \\
            \mathbf{Bernoulli} \left( \sum_{l=1}^K
                \dfrac{l \cdot \mathbf{TruncPoisson}(l; \lambda, K)}{K} \right)
                & \text{$\theta = 0$} \rule{0pt}{4ex} \\
            \mathbf{Bernoulli} \left( \sum_{l=1}^{K-1}
                \dfrac{l \cdot \mathbf{TruncPoisson}(l; \lambda, K-1)}{K-1} \right)
                & \text{otherwise} \rule{0pt}{4ex}
        \end{cases}

    :param lamda: Average rate of target-nonspecific binding.
    :param K: Maximum number of spots that can be present in a single image.
    :return: A tensor of a shape ``lamda.shape + (1 + K, K)`` of probabilities.
    """
    shape = lamda.shape + (1 + K, K)
    dtype = lamda.dtype
    result = torch.zeros(shape, dtype=dtype)
    kdx = torch.arange(K)
    tr_pois_km1 = truncated_poisson_probs(lamda, K - 1)
    km1 = torch.arange(1, K)
    result[..., :, :] = (km1 * tr_pois_km1[..., km1]).sum(-1).unsqueeze(-1).unsqueeze(
        -1
    ) / (K - 1)
    # theta == 0
    tr_pois_k = truncated_poisson_probs(lamda, K)
    k = torch.arange(1, K + 1)
    result[..., 0, :] = (k * tr_pois_k[..., k]).sum(-1).unsqueeze(-1) / K
    # theta == k
    result[..., kdx + 1, kdx] = 1
    return result


def expand_offtarget(probs: torch.Tensor) -> torch.Tensor:
    r"""
    Expand state probability ``probs`` (e.g., :math:`\pi` or :math:`A`) to off-target AOIs.

    .. math::

        p(\mathsf{state}) =
        \begin{cases}
            \mathbf{Categorical}\left( \mathsf{probs} \right) & \textrm{if on-target} \\
            \mathbf{Categorical}\left( \left[ 1, 0, \dots, 0 \right] \right) & \textrm{if off-target}
        \end{cases}

    :param probs: Probability of target-specific states.
    :return: A tensor of a shape ``probs.shape + (2,)`` of probabilities for off-target (``0``)
        and on-target (``1``) AOI.
    """
    offtarget_probs = torch.zeros_like(probs)
    offtarget_probs[..., 0] = 1
    return torch.stack([offtarget_probs, probs], dim=-1)


@lru_cache(maxsize=None)
def probs_theta(K: int, device: torch.device) -> torch.Tensor:
    r"""
    Prior probability for target-specific spot index :math:`p(\theta | z)`.

    .. math::
        p(\theta | z) =
        \begin{cases}
            \mathbf{Categorical}\left( \begin{bmatrix} 0 & 1/K & \dots & 1/K \end{bmatrix} \right) & z > 0 \\
            \mathbf{Categorical}\left( \begin{bmatrix} 1 & 0 & \dots & 0 \end{bmatrix} \right) & z = 0
        \end{cases}

    :param K: Maximum number of spots that can be present in a single image.
    :return: A tensor of a shape ``(2, 1 + K)`` of :math:`\theta` probabilities for spot-absent (``0``)
        and spot-present (``1``) cases.
    """
    result = torch.zeros(2, 1 + K, device=device)
    result[0, 0] = 1
    result[1, 1:] = 1 / K
    return result



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import logging
import random
from collections import deque
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import torch
from pyroapi import infer, optim, pyro
from sklearn.metrics import (
    confusion_matrix,
    matthews_corrcoef,
    precision_score,
    recall_score,
)
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from tapqir import __version__ as tapqir_version
from tapqir.utils.dataset import load
from tapqir.utils.stats import save_stats

logger = logging.getLogger(__name__)


class Model:
    r"""
    Base class for tapqir models.

    Derived models must implement the methods

    * :meth:`model`
    * :meth:`guide`
    * :meth:`init_parameters`
    * :meth:`TraceELBO`

    :param S: Number of distinct molecular states for the binder molecules.
    :param K: Maximum number of spots that can be present in a single image.
    :param channels: Number of color channels.
    :param device: Computation device (cpu or gpu).
    :param dtype: Floating point precision.
    """

    def __init__(
        self,
        S: int = 1,
        K: int = 2,
        channels: Union[tuple, list] = (0,),
        device: str = "cpu",
        dtype: str = "double",
    ):
        self.S = S
        self.K = K
        self.channels = channels
        self.nbatch_size = None
        self.fbatch_size = None
        # for plotting
        self.n = None
        self.f = None
        self.data_path = None
        self.path = None
        self.run_path = None
        # set device & dtype
        self.to(device, dtype)

    def to(self, device: str, dtype: str = "double") -> None:
        """
        Change tensor device and dtype.

        :param device: Computation device, either "gpu" or "cpu".
        :param dtype: Floating point precision, either "double" or "float".
        """
        self.dtype = getattr(torch, dtype)
        self.device = torch.device(device)
        if device == "cuda" and dtype == "double":
            torch.set_default_tensor_type(torch.cuda.DoubleTensor)
        elif device == "cuda" and dtype == "float":
            torch.set_default_tensor_type(torch.cuda.FloatTensor)
        elif device == "cpu" and dtype == "double":
            torch.set_default_tensor_type(torch.DoubleTensor)
        else:
            torch.set_default_tensor_type(torch.FloatTensor)
        # change loaded data device
        if hasattr(self, "data"):
            self.data.ontarget = self.data.ontarget._replace(device=self.device)
            self.data.offtarget = self.data.offtarget._replace(device=self.device)
            self.data.offset = self.data.offset._replace(
                samples=self.data.offset.samples.to(self.device),
                weights=self.data.offset.weights.to(self.device),
            )

    def load(self, path: Union[str, Path], data_only: bool = True) -> None:
        """
        Load data and optionally parameters from a specified path

        :param path: Path to Tapqir analysis folder.
        :param data_only: Load only data or both data and parameters.
        """
        # set path
        self.path = Path(path)
        self.run_path = self.path / ".tapqir"

        # load data
        self.data = load(self.path, self.device)
        logger.debug(f"Loaded data from {self.path / 'data.tpqr'}")

        # load fit results
        if not data_only:
            self.params = torch.load(self.path / f"{self.full_name}-params.tpqr")
            self.summary = pd.read_csv(
                self.path / f"{self.full_name}-summary.csv", index_col=0
            )

    def model(self):
        """
        Generative Model
        """
        raise NotImplementedError

    def guide(self):
        """
        Variational Distribution
        """
        raise NotImplementedError

    def TraceELBO(self, jit):
        """
        A trace implementation of ELBO-based SVI.
        """
        raise NotImplementedError

    def init_parameters(self):
        """
        Initialize variational parameters.
        """
        raise NotImplementedError

    def init(
        self,
        lr: float = 0.005,
        nbatch_size: int = 5,
        fbatch_size: int = 512,
        jit: bool = False,
    ) -> None:
        """
        Initialize SVI object.

        :param lr: Learning rate.
        :param nbatch_size: AOI batch size.
        :param fbatch_size: Frame batch size.
        :param jit: Use JIT compiler.
        """
        self.lr = lr
        self.optim_fn = optim.Adam
        self.optim_args = {"lr": lr, "betas": [0.9, 0.999]}
        self.optim = self.optim_fn(self.optim_args)

        try:
            self.load_checkpoint()
        except (FileNotFoundError, TypeError):
            pyro.clear_param_store()
            self.iter = 0
            self.converged = False
            self._rolling = {p: deque([], maxlen=100) for p in self.conv_params}
            self.init_parameters()

        self.elbo = self.TraceELBO(jit)
        self.svi = infer.SVI(self.model, self.guide, self.optim, loss=self.elbo)

        self.nbatch_size = min(nbatch_size, self.data.Nt)
        self.fbatch_size = min(fbatch_size, self.data.F)

    def run(self, num_iter: int = 0, progress_bar=tqdm) -> int:
        """
        Run inference procedure for a specified number of iterations.
        If num_iter equals zero then run till model converges.

        :param num_iter: Number of iterations.
        :return: Exit code.
        """
        use_crit = False
        if not num_iter:
            use_crit = True
            num_iter = 100000

        logger.debug("Tapqir version - {}".format(tapqir_version))
        logger.debug("Model - {}".format(self.name))
        logger.debug("Device - {}".format(self.device))
        logger.debug("Floating precision - {}".format(self.dtype))
        logger.debug("Optimizer - {}".format(self.optim_fn.__name__))
        logger.debug("Learning rate - {}".format(self.lr))
        logger.debug("AOI batch size - {}".format(self.nbatch_size))
        logger.debug("Frame batch size - {}".format(self.fbatch_size))

        with SummaryWriter(log_dir=self.run_path / "logs" / self.full_name) as writer:
            for i in progress_bar(range(num_iter)):
                try:
                    self.iter_loss = self.svi.step()
                    # save a checkpoint every 200 iterations
                    if not self.iter % 200:
                        self.save_checkpoint(writer)
                        if use_crit and self.converged:
                            logger.debug(f"Iteration #{self.iter} model converged.")
                            return 0
                    self.iter += 1
                except ValueError:
                    # load last checkpoint
                    self.init(
                        lr=self.lr,
                        nbatch_size=self.nbatch_size,
                        fbatch_size=self.fbatch_size,
                    )
                    # change rng seed
                    new_seed = random.randint(0, 100)
                    pyro.set_rng_seed(new_seed)
                    logger.debug(
                        f"Iteration #{self.iter} restarting with a new seed: {new_seed}."
                    )
        return 1

    def save_checkpoint(self, writer: SummaryWriter = None):
        """
        Save checkpoint.

        :param writer: SummaryWriter object.
        """
        # save only if no NaN values
        for k, v in pyro.get_param_store().items():
            if torch.isnan(v).any() or torch.isinf(v).any():
                raise ValueError(
                    "Iteration #{}. Detected NaN values in {}".format(self.iter, k)
                )

        # update convergence criteria parameters
        for name in self.conv_params:
            if name == "-ELBO":
                self._rolling["-ELBO"].append(self.iter_loss)
            else:
                self._rolling[name].append(pyro.param(name).item())

        # check convergence status
        self.converged = False
        if len(self._rolling["-ELBO"]) == self._rolling["-ELBO"].maxlen:
            crit = all(
                torch.tensor(self._rolling[p]).std()
                / torch.tensor(self._rolling[p])[-50:].std()
                < 1.05
                for p in self.conv_params
            )
            if crit:
                self.converged = True

        # save the model state
        torch.save(
            {
                "iter": self.iter,
                "params": pyro.get_param_store().get_state(),
                "optimizer": self.optim.get_state(),
                "rolling": self._rolling,
                "convergence_status": self.converged,
            },
            self.run_path / f"{self.full_name}-model.tpqr",
        )

        # save global paramters for tensorboard
        writer.add_scalar("-ELBO", self.iter_loss, self.iter)
        for name, val in pyro.get_param_store().items():
            if val.dim() == 0:
                writer.add_scalar(name, val.item(), self.iter)
            elif val.dim() == 1 and len(val) <= self.S + 1:
                scalars = {str(i): v.item() for i, v in enumerate(val)}
                writer.add_scalars(name, scalars, self.iter)

        if False and self.data.labels is not None:
            pred_labels = (
                self.pspecific_map[self.data.is_ontarget].cpu().numpy().ravel()
            )
            true_labels = self.data.labels["z"].ravel()

            metrics = {}
            with np.errstate(divide="ignore", invalid="ignore"):
                metrics["MCC"] = matthews_corrcoef(true_labels, pred_labels)
            metrics["Recall"] = recall_score(true_labels, pred_labels, zero_division=0)
            metrics["Precision"] = precision_score(
                true_labels, pred_labels, zero_division=0
            )

            neg, pos = {}, {}
            neg["TN"], neg["FP"], pos["FN"], pos["TP"] = confusion_matrix(
                true_labels, pred_labels, labels=(0, 1)
            ).ravel()

            writer.add_scalars("ACCURACY", metrics, self.iter)
            writer.add_scalars("NEGATIVES", neg, self.iter)
            writer.add_scalars("POSITIVES", pos, self.iter)

        logger.debug(f"Iteration #{self.iter}: Successful.")

    def load_checkpoint(
        self,
        path: Union[str, Path] = None,
        param_only: bool = False,
        warnings: bool = False,
    ):
        """
        Load checkpoint.

        :param path: Path to model checkpoint.
        :param param_only: Load only parameters.
        :param warnings: Give warnings if loaded model has not been fully trained.
        """
        device = self.device
        path = Path(path) if path else self.run_path
        pyro.clear_param_store()
        checkpoint = torch.load(
            path / f"{self.full_name}-model.tpqr", map_location=device
        )
        pyro.get_param_store().set_state(checkpoint["params"])
        if not param_only:
            self.converged = checkpoint["convergence_status"]
            self._rolling = checkpoint["rolling"]
            self.iter = checkpoint["iter"]
            self.optim.set_state(checkpoint["optimizer"])
            logger.info(
                f"Iteration #{self.iter}. Loaded a model checkpoint from {path}"
            )
        if warnings and not checkpoint["convergence_status"]:
            logger.warning(f"Model at {path} has not been fully trained")

    def compute_stats(self, CI: float = 0.95, save_matlab: bool = False):
        """
        Compute credible regions (CI) and other stats.

        :param CI: credible region.
        :param save_matlab: Save output in Matlab format as well.
        """
        save_stats(self, self.path, CI=CI, save_matlab=save_matlab)
        if self.path is not None:
            logger.info(f"Parameters were saved in {self.path / 'params.tpqr'}")
            if save_matlab:
                logger.info(f"Parameters were saved in {self.path / 'params.mat'}")



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import numpy as np
import torch
from pyro.infer import Predictive
from pyroapi import handlers, pyro

from tapqir.utils.dataset import CosmosDataset


def simulate(
    model: str,
    N: int,
    F: int,
    C: int = 1,
    P: int = 14,
    seed: int = 0,
    params: dict = dict(),
) -> CosmosDataset:
    """
    Simulate a new dataset.

    :param model: Tapqir model.
    :param N: Number of total AOIs. Half will be on-target and half off-target.
    :param F: Number of frames.
    :param C: Number of color channels.
    :param P: Number of pixels alongs the axis.
    :param seed: Rng seed.
    :param params: A dictionary of fixed parameter values.
    :return: A new simulated data set.
    """

    pyro.set_rng_seed(seed)
    pyro.clear_param_store()

    # samples
    samples = {}
    samples["gain"] = torch.full((1, 1), params["gain"])
    samples["lamda"] = torch.full((1, 1), params["lamda"])
    samples["proximity"] = torch.full((1, 1), params["proximity"])

    if "pi" in params:
        samples["pi"] = torch.tensor([[1 - params["pi"], params["pi"]]])
        samples["background"] = torch.full((1, N, 1), params["background"])
        for k in range(model.K):
            samples[f"width_{k}"] = torch.full((1, N, F), params["width"])
            samples[f"height_{k}"] = torch.full((1, N, F), params["height"])
    else:
        # kinetic simulations
        samples["init"] = torch.tensor(
            [
                [
                    params["koff"] / (params["kon"] + params["koff"]),
                    params["kon"] / (params["kon"] + params["koff"]),
                ]
            ]
        )
        samples["trans"] = torch.tensor(
            [
                [
                    [1 - params["kon"], params["kon"]],
                    [params["koff"], 1 - params["koff"]],
                ]
            ]
        )
        for f in range(F):
            samples[f"background_{f}"] = torch.full((1, N, 1), params["background"])
            for k in range(model.K):
                samples[f"width_{k}_{f}"] = torch.full((1, N, 1), params["width"])
                samples[f"height_{k}_{f}"] = torch.full((1, N, 1), params["height"])

    offset = torch.full((3,), params["offset"])
    target_locs = torch.full((N, F, 1, 2), (P - 1) / 2)
    is_ontarget = torch.zeros((N,), dtype=torch.bool)
    is_ontarget[: N // 2] = True
    # placeholder dataset
    model.data = CosmosDataset(
        torch.full((N, F, C, P, P), params["background"] + params["offset"]),
        target_locs,
        is_ontarget,
        None,
        offset_samples=offset,
        offset_weights=torch.ones(3) / 3,
        device=model.device,
    )

    # sample
    predictive = Predictive(
        handlers.uncondition(model.model), posterior_samples=samples, num_samples=1
    )
    samples = predictive()
    data = torch.zeros(N, F, C, P, P)
    labels = np.zeros((N // 2, F, 1), dtype=[("aoi", int), ("frame", int), ("z", bool)])
    labels["aoi"] = np.arange(N // 2).reshape(-1, 1, 1)
    labels["frame"] = np.arange(F).reshape(-1, 1)
    if "pi" in params:
        data[:, :, 0] = samples["data"][0].data.floor()
        labels["z"][:, :, 0] = samples["theta"][0][: N // 2].cpu() > 0
    else:
        # kinetic simulations
        for f in range(F):
            data[:, f : f + 1, 0] = samples[f"data_{f}"][0].data.floor()
            labels["z"][:, f : f + 1, 0] = samples[f"theta_{f}"][0][: N // 2].cpu() > 0

    return CosmosDataset(
        data.cpu(),
        target_locs.cpu(),
        is_ontarget.cpu(),
        labels,
        offset_samples=offset,
        offset_weights=torch.ones(3) / 3,
        device=model.device,
    )



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import os
from collections import defaultdict
from functools import partial, singledispatch
from pathlib import Path

import ipywidgets as widgets
import matplotlib.pyplot as plt
import torch
import typer
from ipyevents import Event
from ipyfilechooser import FileChooser
from tensorboard import notebook
from tqdm import tqdm_notebook
from traitlets.utils.bunch import Bunch

from tapqir.distributions.util import gaussian_spots
from tapqir.main import fit, glimpse, main, show

# disable pyplot keyboard shortcuts
plt.rcParams["keymap.zoom"].remove("o")
plt.rcParams["keymap.home"].remove("h")
plt.rcParams["keymap.yscale"].remove("l")
plt.rcParams["keymap.xscale"].remove("k")


@singledispatch
def get_value(x):
    return x


@get_value.register
def _(x: Bunch):
    return x.new


@get_value.register
def _(x: widgets.Widget):
    return x.value


@singledispatch
def get_path(x):
    return x


@get_path.register
def _(x: str):
    return Path(x)


@get_path.register
def _(x: FileChooser):
    return Path(x.selected_path)


def initUI(DEFAULTS):
    """
    Main layout.

    - Set working directory.
    - Commands tabs.
    - Output.
    """

    layout = widgets.VBox(layout={"width": "850px", "border": "2px solid blue"})

    # widgets
    cd = FileChooser(".", title="Select working directory")
    cd.show_only_dirs = True

    out = widgets.Output(layout={"border": "1px solid black"})

    # layout
    layout.children = [cd, out]
    # callbacks
    cd.register_callback(
        partial(
            cdCmd,
            DEFAULTS=DEFAULTS,
            out=out,
            layout=layout,
        )
    )
    return layout


def cdCmd(path, DEFAULTS, out, layout):
    """
    Set working directory and load default parameters (main).
    """
    import sys

    IN_COLAB = "google.colab" in sys.modules

    path = get_path(path)
    with out:
        typer.echo("Loading configuration data ...")
        main(cd=path)

    # Tabs
    tab = widgets.Tab()
    tab.children = [glimpseUI(out), fitUI(out)]
    if not IN_COLAB:
        tensorboard = widgets.Output(layout={"border": "1px solid blue"})
        tab.children = tab.children + (showUI(out), tensorboard)
    else:
        tensorboard = None
        tab.children = tab.children + (
            widgets.Label(value="Disabled in Colab"),
            widgets.Label(value="Disabled in Colab"),
        )
    tab.set_title(0, "Extract AOIs")
    tab.set_title(1, "Fit the data")
    tab.set_title(2, "View results")
    tab.set_title(3, "Tensorboard")

    if tensorboard is not None:
        with tensorboard:
            notebook.start(f"--logdir '{path}'")
            notebook.display(height=1000)
    glimpseTab = tab.children[0]
    for i, flag in enumerate(
        [
            "dataset",
            "P",
            "frame-range",
            "frame-start",
            "frame-end",
            "num-channels",
            "use-offtarget",
        ]
    ):
        if DEFAULTS[flag] is not None:
            glimpseTab.children[i].value = DEFAULTS[flag]
    channelTabs = glimpseTab.children[7]
    for c, channel in enumerate(channelTabs.children):
        if len(DEFAULTS["channels"]) < c + 1:
            DEFAULTS["channels"].append(defaultdict(lambda: None))
        for widget, flag in zip(
            channel.children,
            [
                "name",
                "glimpse-folder",
                "driftlist",
                "ontarget-aoiinfo",
                "offtarget-aoiinfo",
            ],
        ):
            if DEFAULTS["channels"][c][flag] is not None:
                if isinstance(widget, FileChooser):
                    if flag == "glimpse-folder":
                        widget.reset(path=Path(DEFAULTS["channels"][c][flag]))
                    else:
                        widget.reset(
                            path=Path(DEFAULTS["channels"][c][flag]).parent,
                            filename=Path(DEFAULTS["channels"][c][flag]).name,
                        )
                    widget._apply_selection()
                elif flag:
                    widget.value = DEFAULTS["channels"][c][flag]
    numChannels = glimpseTab.children[5].value
    fitTab = tab.children[1]
    fitTab.children[1].options = [str(c) for c in range(numChannels)]
    if not IN_COLAB:
        showTab = tab.children[2]
        showTab.children[1].options = [str(c) for c in range(numChannels)]
    for i, flag in enumerate(
        [
            False,
            False,
            "cuda",
            "nbatch-size",
            "fbatch-size",
            "learning-rate",
            False,
            "matlab",
        ]
    ):
        if flag and DEFAULTS[flag] is not None:
            fitTab.children[i].value = DEFAULTS[flag]

    # insert tabs into GUI
    wd = widgets.Label(value=f"Working directory: {path}")
    layout.children = (wd, tab) + layout.children[1:]

    with out:
        typer.echo("Loading configuration data: Done")

    out.clear_output(wait=True)


def glimpseUI(out):
    """
    Glimpse command layout.
    """
    # Extract AOIs
    layout = widgets.VBox()
    # Dataset name
    datasetName = widgets.Text(
        description="Dataset name",
        style={"description_width": "initial"},
    )
    # AOI image size
    aoiSize = widgets.IntText(
        description="AOI image size",
        style={"description_width": "initial"},
    )
    # Specify frame range?
    specifyFrame = widgets.Checkbox(
        value=False, description="Specify frame range?", indent=False
    )
    # First frame
    firstFrame = widgets.IntText(
        description="Starting frame",
        style={"description_width": "initial"},
    )
    # Last frame
    lastFrame = widgets.IntText(
        description="Ending frame",
        style={"description_width": "initial"},
    )
    # Number of channels
    numChannels = widgets.BoundedIntText(
        min=1,
        max=4,
        step=1,
        description="Number of color channels",
        style={"description_width": "initial"},
    )
    # Specify frame range?
    useOfftarget = widgets.Checkbox(
        description="Use off-target AOI locations?",
        indent=False,
    )
    # Channel tabs
    channelTabs = widgets.Tab()
    channelUI(numChannels.value, channelTabs, useOfftarget)
    # extract AOIs
    extractAOIs = widgets.Button(description="Extract AOIs")
    # Layout
    layout.children = [
        datasetName,
        aoiSize,
        specifyFrame,
        firstFrame,
        lastFrame,
        numChannels,
        useOfftarget,
        channelTabs,
        extractAOIs,
    ]
    # Callbacks
    numChannels.observe(
        partial(channelUI, channelTabs=channelTabs, useOfftarget=useOfftarget),
        names="value",
    )
    useOfftarget.observe(
        partial(toggleUseOffTarget, channelTabs=channelTabs), names="value"
    )
    extractAOIs.on_click(partial(glimpseCmd, layout=layout, out=out))
    specifyFrame.observe(
        partial(toggleWidgets, widgets=(firstFrame, lastFrame)), names="value"
    )
    return layout


def toggleUseOffTarget(checked, channelTabs):
    checked = get_value(checked)
    if checked:
        for tab in channelTabs.children:
            offtargetLayout = FileChooser(
                ".", title="Off-target control locations file"
            )
            tab.children = tab.children + (offtargetLayout,)
    else:
        for tab in channelTabs.children:
            tab.children = tab.children[:-1]


def glimpseCmd(b, layout, out):
    channelTabs = layout.children[7]
    with out:
        glimpse(
            dataset=layout.children[0].value,
            P=layout.children[1].value,
            frame_range=layout.children[2].value,
            frame_start=layout.children[3].value,
            frame_end=layout.children[4].value,
            num_channels=layout.children[5].value,
            use_offtarget=layout.children[6].value,
            name=[c.children[0].value for c in channelTabs.children],
            glimpse_folder=[c.children[1].value for c in channelTabs.children],
            ontarget_aoiinfo=[c.children[3].value for c in channelTabs.children],
            offtarget_aoiinfo=[c.children[4].value for c in channelTabs.children],
            driftlist=[c.children[2].value for c in channelTabs.children],
            no_input=True,
            progress_bar=tqdm_notebook,
            labels=False,
        )

    out.clear_output(wait=True)


def channelUI(C, channelTabs, useOfftarget):
    C = get_value(C)
    currentC = len(channelTabs.children)
    for i in range(max(currentC, C)):
        if i < C and i < currentC:
            continue
        elif i < C and i >= currentC:
            layout = widgets.VBox()
            nameLayout = widgets.Text(
                description="Channel name", style={"description_width": "initial"}
            )
            headerLayout = FileChooser(".", title="Header/glimpse folder")
            ontargetLayout = FileChooser(".", title="Target molecule locations file")
            driftlistLayout = FileChooser(".", title="Driftlist file")
            layout.children = [
                nameLayout,
                headerLayout,
                driftlistLayout,
                ontargetLayout,
            ]
            if useOfftarget.value:
                offtargetLayout = FileChooser(
                    ".", title="Off-target control locations file"
                )
                layout.children = layout.children + (offtargetLayout,)
            channelTabs.children = channelTabs.children + (layout,)
        channelTabs.set_title(i, f"Channel #{i}")
    channelTabs.children = channelTabs.children[:C]


def fitUI(out):
    # Fit the data
    layout = widgets.VBox()
    # Model
    modelSelect = widgets.Dropdown(
        description="Tapqir model",
        value="cosmos",
        options=["cosmos"],
        style={"description_width": "initial"},
    )
    # Channel numbers
    channelNumber = widgets.Dropdown(
        description="Channel numbers",
        value="0",
        options=["0"],
        style={"description_width": "initial"},
    )
    # Run computations on GPU?
    useGpu = widgets.Checkbox(
        value=True,
        description="Run computations on GPU?",
        style={"description_width": "initial"},
    )
    # AOI batch size
    aoiBatch = widgets.IntText(
        description="AOI batch size",
        style={"description_width": "initial"},
    )
    # Frame batch size
    frameBatch = widgets.IntText(
        description="Frame batch size",
        style={"description_width": "initial"},
    )
    # Learning rate
    learningRate = widgets.FloatText(
        description="Learning rate",
        style={"description_width": "initial"},
    )
    # Number of iterations
    iterationNumber = widgets.IntText(
        description="Number of iterations",
        style={"description_width": "initial"},
    )
    # Save parameters in matlab format?
    saveMatlab = widgets.Checkbox(
        value=False,
        description="Save parameters in matlab format?",
        style={"description_width": "initial"},
    )
    # Fit the data
    fitData = widgets.Button(description="Fit the data")
    # Layout
    layout.children = [
        modelSelect,
        channelNumber,
        useGpu,
        aoiBatch,
        frameBatch,
        learningRate,
        iterationNumber,
        saveMatlab,
        fitData,
    ]
    # Callbacks
    fitData.on_click(partial(fitCmd, layout=layout, out=out))
    return layout


def fitCmd(b, layout, out):
    with out:
        fit(
            model=layout.children[0].value,
            channels=[int(layout.children[1].value)],
            cuda=layout.children[2].value,
            nbatch_size=layout.children[3].value,
            fbatch_size=layout.children[4].value,
            learning_rate=layout.children[5].value,
            num_iter=layout.children[6].value,
            matlab=layout.children[7].value,
            k_max=2,
            funsor=False,
            pykeops=True,
            no_input=True,
            progress_bar=tqdm_notebook,
        )

    out.clear_output(wait=True)


def showUI(out):
    # Fit the data
    layout = widgets.VBox()
    # Model
    modelSelect = widgets.Dropdown(
        description="Tapqir model",
        value="cosmos",
        options=["cosmos"],
        style={"description_width": "initial"},
    )
    # Channel numbers
    channelNumber = widgets.Dropdown(
        description="Channel numbers",
        value="0",
        options=["0"],
        style={"description_width": "initial"},
    )
    # Fit the data
    showParams = widgets.Button(description="Load results")
    controls = widgets.HBox(layout={"border": "1px solid grey"})
    view = widgets.Output(layout={"border": "1px solid red"})
    # Layout
    layout.children = [
        modelSelect,
        channelNumber,
        showParams,
        controls,
        view,
    ]
    # Callbacks
    showParams.on_click(partial(showCmd, layout=layout, view=view, out=out))
    return layout


def showCmd(b, layout, view, out):
    with out:
        typer.echo("Loading results ...")
    with view:
        model, fig, item, ax = show(
            model=layout.children[0].value,
            channels=[int(layout.children[1].value)],
            n=0,
            f1=None,
            f2=None,
        )
    controls = layout.children[3]
    n = widgets.BoundedIntText(
        value=0,
        min=0,
        max=model.data.Nt - 1,
        description="AOI",
        style={"description_width": "initial"},
        layout={"width": "150px"},
    )
    n_counter = widgets.BoundedIntText(
        min=0,
        max=model.data.Nt - 1,
    )
    f1_text = widgets.BoundedIntText(
        value=0,
        min=0,
        max=model.data.F - 15,
        step=1,
        description="Frame",
        style={"description_width": "initial"},
        layout={"width": "300px"},
    )
    f1_counter = widgets.BoundedIntText(
        min=0,
        max=model.data.F - 15,
    )
    f1_slider = widgets.IntSlider(
        value=0,
        min=0,
        max=model.data.F - 15,
        step=1,
        continuous_update=True,
        readout=False,
        readout_format="d",
        layout={"width": "210px"},
    )
    f1_box = widgets.VBox(layout=widgets.Layout(width="310px"))
    f1_incr = widgets.Button(description="+15", layout=widgets.Layout(width="45px"))
    f1_decr = widgets.Button(description="-15", layout=widgets.Layout(width="45px"))
    f1_controls = widgets.HBox(
        children=[f1_decr, f1_slider, f1_incr],
        layout=widgets.Layout(width="305px"),
    )
    f1_box.children = [f1_text, f1_controls]
    widgets.jslink((f1_slider, "value"), (f1_text, "value"))
    widgets.jslink((f1_text, "value"), (f1_counter, "value"))
    widgets.jslink((n, "value"), (n_counter, "value"))
    zoom = widgets.Checkbox(
        value=False,
        description="Zoom out frames ['z']",
        indent=False,
        layout={"width": "240px"},
    )
    labels = widgets.Checkbox(
        value=False,
        description="Show labels",
        indent=False,
        layout={"width": "240px"},
    )
    targets = widgets.Checkbox(
        value=False,
        description="Show target location ['o']",
        indent=False,
        layout={"width": "240px"},
    )
    nonspecific = widgets.Checkbox(
        value=True,
        description="Show non-specific spots ['n']",
        indent=False,
        layout={"width": "240px"},
    )
    checkboxes = widgets.VBox(layout=widgets.Layout(width="250px"))
    if model.data.labels is None:
        checkboxes.children = [zoom, targets, nonspecific]
    else:
        checkboxes.children = [zoom, targets, nonspecific, labels]
    controls.children = [n, f1_box, checkboxes]
    n.observe(
        partial(
            updateParams,
            f1=f1_slider,
            model=model,
            fig=fig,
            item=item,
            ax=ax,
            targets=targets,
            nonspecific=nonspecific,
            labels=labels,
        ),
        names="value",
    )
    f1_slider.observe(
        partial(
            updateRange,
            n=n,
            model=model,
            fig=fig,
            item=item,
            ax=ax,
            zoom=zoom,
            targets=targets,
        ),
        names="value",
    )
    f1_incr.on_click(partial(incrementRange, x=15, counter=f1_counter))
    f1_decr.on_click(partial(incrementRange, x=-15, counter=f1_counter))
    zoom.observe(
        partial(zoomOut, f1=f1_slider, model=model, fig=fig, item=item, ax=ax),
        names="value",
    )
    labels.observe(
        partial(showLabels, n=n_counter, model=model, item=item, ax=ax),
        names="value",
    )
    targets.observe(
        partial(showTargets, n=n_counter, f1=f1_slider, model=model, item=item, ax=ax),
        names="value",
    )
    nonspecific.observe(
        partial(showNonspecific, n=n_counter, model=model, item=item, ax=ax),
        names="value",
    )
    # mouse click UI
    fig.canvas.mpl_connect(
        "button_release_event",
        partial(onFrameClick, counter=f1_counter),
    )
    # key press UI
    d = Event(source=layout, watched_events=["keyup"], prevent_default_action=True)
    d.on_dom_event(
        partial(
            onKeyPress,
            n=n_counter,
            f1=f1_counter,
            zoom=zoom,
            targets=targets,
            nonspecific=nonspecific,
        )
    )
    with out:
        typer.echo("Loading results: Done")

    out.clear_output(wait=True)


def onKeyPress(event, n, f1, zoom, targets, nonspecific):
    key = event["key"]
    if key == "h" or key == "ArrowLeft":
        f1.value = f1.value - 15
    elif key == "l" or key == "ArrowRight":
        f1.value = f1.value + 15
    elif key == "j" or key == "ArrowDown":
        n.value = n.value - 1
    elif key == "k" or key == "ArrowUp":
        n.value = n.value + 1
    elif key == "z":
        zoom.value = not zoom.value
    elif key == "o":
        targets.value = not targets.value
    elif key == "n":
        nonspecific.value = not nonspecific.value


def onFrameClick(event, counter):
    counter.value = int(event.xdata)


def incrementRange(name, x, counter):
    counter.value = counter.value + x


def updateParams(n, f1, model, fig, item, ax, targets, nonspecific, labels):
    n = get_value(n)
    f1 = get_value(f1)
    f2 = f1 + 15
    c = model.cdx

    frames = torch.arange(f1, f2)
    img_ideal = (
        model.data.offset.mean
        + model.params["background"]["Mean"][n, frames, None, None]
    )
    gaussian = gaussian_spots(
        model.params["height"]["Mean"][:, n, frames],
        model.params["width"]["Mean"][:, n, frames],
        model.params["x"]["Mean"][:, n, frames],
        model.params["y"]["Mean"][:, n, frames],
        model.data.xy[n, frames, c],
        model.data.P,
    )
    img_ideal = img_ideal + gaussian.sum(-4)
    for i, f in enumerate(range(f1, f2)):
        ax[f"image_{i}"].set_title(rf"${f}$", fontsize=9)
        item[f"image_{i}"].set_data(model.data.images[n, f, c].numpy())
        item[f"ideal_{i}"].set_data(img_ideal[i].numpy())
        if targets.value:
            item[f"target_{i}"].remove()
            item[f"target_{i}"] = ax[f"image_{i}"].scatter(
                model.data.x[n, f, c].item(),
                model.data.y[n, f, c].item(),
                c="C0",
                s=40,
                marker="+",
            )

    params = ["p_specific", "z_map", "height", "width", "x", "y", "background", "chi2"]
    if labels.value:
        params += ["labels"]
    theta_mask = model.params["theta_probs"][:, n] > 0.5
    j_mask = (model.params["m_probs"][:, n] > 0.5) & ~theta_mask
    for p in params:
        if p == "p_specific":
            item[p].set_ydata(model.params[p][n])
        elif p in {"z_map"}.intersection(model.params.keys()):
            item[p].set_ydata(model.params[p][n])
        elif p == "chi2":
            item[p].set_ydata(model.params[p]["values"][n])
        elif p == "background":
            item[f"{p}_fill"].remove()
            item[f"{p}_fill"] = ax[p].fill_between(
                torch.arange(0, model.data.F),
                model.params[p]["LL"][n],
                model.params[p]["UL"][n],
                alpha=0.3,
                color="k",
            )
            item[f"{p}_mean"].set_ydata(model.params[p]["Mean"][n])
        elif p == "labels":
            item[p].set_ydata(model.data.labels["z"][n, :, c])
        elif p in ["height", "width", "x", "y"]:
            # target-nonspecific spots
            if nonspecific.value:
                for k in range(model.K):
                    f_mask = j_mask[k]
                    mean = model.params[p]["Mean"][k, n] * f_mask
                    ll = model.params[p]["LL"][k, n] * f_mask
                    ul = model.params[p]["UL"][k, n] * f_mask
                    item[f"{p}_nonspecific{k}_mean"].remove()
                    (item[f"{p}_nonspecific{k}_mean"],) = ax[p].plot(
                        torch.arange(0, model.data.F)[f_mask],
                        mean[f_mask],
                        "o",
                        ms=2,
                        lw=1,
                        color=f"C{k}",
                    )
                    item[f"{p}_nonspecific{k}_fill"].remove()
                    item[f"{p}_nonspecific{k}_fill"] = ax[p].fill_between(
                        torch.arange(0, model.data.F),
                        ll,
                        ul,
                        where=f_mask,
                        alpha=0.3,
                        color=f"C{k}",
                    )
            # target-specific spots
            f_mask = theta_mask.sum(0).bool()
            mean = (model.params[p]["Mean"][:, n] * theta_mask).sum(0)
            ll = (model.params[p]["LL"][:, n] * theta_mask).sum(0)
            ul = (model.params[p]["UL"][:, n] * theta_mask).sum(0)
            item[f"{p}_specific_fill"].remove()
            item[f"{p}_specific_fill"] = ax[p].fill_between(
                torch.arange(0, model.data.F),
                ll,
                ul,
                where=f_mask,
                alpha=0.3,
                color="C2",
            )
            item[f"{p}_specific_mean"].remove()
            (item[f"{p}_specific_mean"],) = ax[p].plot(
                torch.arange(0, model.data.F)
                if p == "height"
                else torch.arange(0, model.data.F)[f_mask],
                mean if p == "height" else mean[f_mask],
                "-" if p == "height" else "o",
                ms=2,
                color="C2",
            )
    fig.canvas.draw()


def updateRange(f1, n, model, fig, item, ax, zoom, targets):
    n = get_value(n)
    f1 = get_value(f1)
    f2 = f1 + 15
    c = model.cdx

    frames = torch.arange(f1, f2)
    img_ideal = (
        model.data.offset.mean
        + model.params["background"]["Mean"][n, frames, None, None]
    )
    gaussian = gaussian_spots(
        model.params["height"]["Mean"][:, n, frames],
        model.params["width"]["Mean"][:, n, frames],
        model.params["x"]["Mean"][:, n, frames],
        model.params["y"]["Mean"][:, n, frames],
        model.data.xy[n, frames, c],
        model.data.P,
    )
    img_ideal = img_ideal + gaussian.sum(-4)
    for i, f in enumerate(range(f1, f2)):
        ax[f"image_{i}"].set_title(rf"${f}$", fontsize=9)
        item[f"image_{i}"].set_data(model.data.images[n, f, c].numpy())
        item[f"ideal_{i}"].set_data(img_ideal[i].numpy())
        if targets.value:
            item[f"target_{i}"].remove()
            item[f"target_{i}"] = ax[f"image_{i}"].scatter(
                model.data.x[n, f, c].item(),
                model.data.y[n, f, c].item(),
                c="C0",
                s=40,
                marker="+",
            )

    if not zoom.value:
        for key, a in ax.items():
            if not key.startswith("image") and not key.startswith("ideal"):
                a.set_xlim(f1 - 0.5, f2 - 0.5)
    else:
        for p in [
            "z_map",
            "p_specific",
            "height",
            "width",
            "x",
            "y",
            "background",
            "chi2",
        ]:
            item[f"{p}_vspan"].remove()
            item[f"{p}_vspan"] = ax[p].axvspan(f1, f2, facecolor="C0", alpha=0.3)
    fig.canvas.draw()


def zoomOut(checked, f1, model, fig, item, ax):
    checked = get_value(checked)
    f1_value = get_value(f1)
    if checked:
        f1 = 0
        f2 = model.data.F
        for p in [
            "z_map",
            "p_specific",
            "height",
            "width",
            "x",
            "y",
            "background",
            "chi2",
        ]:
            item[f"{p}_vspan"] = ax[p].axvspan(
                f1_value, f1_value + 15, facecolor="C0", alpha=0.3
            )
    else:
        f1 = f1_value
        f2 = f1 + 15
        for p in [
            "z_map",
            "p_specific",
            "height",
            "width",
            "x",
            "y",
            "background",
            "chi2",
        ]:
            item[f"{p}_vspan"].remove()
    for key, a in ax.items():
        if not key.startswith("image") and not key.startswith("ideal"):
            a.set_xlim(f1 - 0.5, f2 - 0.5)
    fig.canvas.draw()


def showLabels(checked, n, model, item, ax):
    checked = get_value(checked)
    n = get_value(n)
    if checked:
        (item["labels"],) = ax["z_map"].plot(
            torch.arange(0, model.data.F),
            model.data.labels["z"][n, :, model.cdx],
            "-",
            lw=1,
            color="r",
        )
    else:
        item["labels"].remove()


def showTargets(checked, n, f1, model, item, ax):
    checked = get_value(checked)
    n = get_value(n)
    f1 = get_value(f1)
    f2 = f1 + 15
    c = model.cdx
    if checked:
        for i, f in enumerate(range(f1, f2)):
            item[f"target_{i}"] = ax[f"image_{i}"].scatter(
                model.data.x[n, f, c].item(),
                model.data.y[n, f, c].item(),
                c="C0",
                s=40,
                marker="+",
            )
    else:
        for i, f in enumerate(range(f1, f2)):
            item[f"target_{i}"].remove()


def showNonspecific(checked, n, model, item, ax):
    checked = get_value(checked)
    n = get_value(n)
    for p in ["height", "width", "x", "y"]:
        if checked:
            theta_mask = model.params["theta_probs"][:, n] > 0.5
            j_mask = (model.params["m_probs"][:, n] > 0.5) & ~theta_mask
            # target-nonspecific spots
            for k in range(model.K):
                f_mask = j_mask[k]
                mean = model.params[p]["Mean"][k, n] * f_mask
                ll = model.params[p]["LL"][k, n] * f_mask
                ul = model.params[p]["UL"][k, n] * f_mask
                (item[f"{p}_nonspecific{k}_mean"],) = ax[p].plot(
                    torch.arange(0, model.data.F)[f_mask],
                    mean[f_mask],
                    "o",
                    ms=2,
                    lw=1,
                    color=f"C{k}",
                )
                item[f"{p}_nonspecific{k}_fill"] = ax[p].fill_between(
                    torch.arange(0, model.data.F),
                    ll,
                    ul,
                    where=f_mask,
                    alpha=0.3,
                    color=f"C{k}",
                )
        else:
            for k in range(model.K):
                item[f"{p}_nonspecific{k}_fill"].remove()
                item[f"{p}_nonspecific{k}_mean"].remove()


def toggleWidgets(b, widgets):
    checked = get_value(b)
    for w in widgets:
        w.disabled = not checked


def app():
    import pkg_resources

    nbpath = pkg_resources.resource_filename("tapqir", "app.ipynb")
    os.system("voila " + nbpath)


def run():
    from IPython.display import display

    from tapqir.main import DEFAULTS

    display(initUI(DEFAULTS))



# Copyright Contributors to the Tapqir project.
# SPDX-License-Identifier: Apache-2.0

import torch
import torch.distributions.constraints as constraints
from pyro.ops.indexing import Vindex
from pyroapi import distributions as dist
from pyroapi import handlers, infer, optim, pyro


def train(model, guide, lr=1e-3, n_steps=1000, jit=True, verbose=False, **kwargs):

    pyro.clear_param_store()
    optimizer = optim.Adam({"lr": lr})
    elbo = (
        infer.JitTraceEnum_ELBO(max_plate_nesting=2)
        if jit
        else infer.TraceEnum_ELBO(max_plate_nesting=2)
    )
    svi = infer.SVI(model, guide, optimizer, elbo)

    for step in range(n_steps):
        svi.step(**kwargs)
        if step % 100 == 99 and verbose:
            values = tuple(f"{k}: {v}" for k, v in pyro.get_param_store().items())
            print(values)


def ttfb_model(data, control, Tmax):
    r"""
    Eq. 4 and Eq. 7 in::

      @article{friedman2015multi,
        title={Multi-wavelength single-molecule fluorescence analysis of transcription mechanisms},
        author={Friedman, Larry J and Gelles, Jeff},
        journal={Methods},
        volume={86},
        pages={27--36},
        year={2015},
        publisher={Elsevier}
      }

    :param data: time prior to the first binding at the target location
    :param control: time prior to the first binding at the control location
    :param Tmax: entire observation interval
    """
    ka = pyro.param(
        "ka",
        lambda: torch.full((data.shape[0], 1), 0.001),
        constraint=constraints.positive,
    )
    kns = pyro.param(
        "kns",
        lambda: torch.full((data.shape[0], 1), 0.001),
        constraint=constraints.positive,
    )
    Af = pyro.param(
        "Af",
        lambda: torch.full((data.shape[0], 1), 0.9),
        constraint=constraints.unit_interval,
    )
    k = torch.stack([kns, ka + kns])

    # on-target data
    mask = (data < Tmax) & (data > 0)
    tau = data.masked_fill(~mask, 1.0)
    with pyro.plate("bootstrap", data.shape[0], dim=-2) as bdx:
        with pyro.plate("N", data.shape[1], dim=-1):
            active = pyro.sample(
                "active", dist.Bernoulli(Af), infer={"enumerate": "parallel"}
            )
            with handlers.mask(mask=(data == Tmax)):
                pyro.factor("Tmax", -Vindex(k)[active.long().squeeze(-1), bdx] * Tmax)
                # pyro.factor("Tmax", -k * Tmax)
            with handlers.mask(mask=mask):
                pyro.sample(
                    "tau",
                    dist.Exponential(Vindex(k)[active.long().squeeze(-1), bdx]),
                    obs=tau,
                )
                # pyro.sample("tau", dist.Exponential(k), obs=tau)

    # negative control data
    if control is not None:
        mask = (control < Tmax) & (control > 0)
        tauc = control.masked_fill(~mask, 1.0)
        with pyro.plate("bootstrapc", control.shape[0], dim=-2):
            with pyro.plate("Nc", control.shape[1], dim=-1):
                with handlers.mask(mask=(control == Tmax)):
                    pyro.factor("Tmaxc", -kns * Tmax)
                with handlers.mask(mask=mask):
                    pyro.sample("tauc", dist.Exponential(kns), obs=tauc)


def ttfb_guide(data, control, Tmax):
    pass  # MLE


def double_exp_model(data):
    k1 = pyro.param("k1", lambda: torch.tensor(0.01), constraint=constraints.positive)
    k2 = pyro.param("k2", lambda: torch.tensor(0.05), constraint=constraints.positive)
    A = pyro.param("A", lambda: torch.tensor(0.5), constraint=constraints.unit_interval)
    k = torch.stack([k1, k2])

    with pyro.plate("data", len(data)):
        m = pyro.sample("m", dist.Bernoulli(A), infer={"enumerate": "parallel"})
        pyro.sample("obs", dist.Exponential(k[m.long()]), obs=data)


def double_exp_guide(data):
    pass  # MLE
